import torch
import os
from IPython import embed


def XTransE_predict(batch, model, predicion="all"):
    if predicion == "all":
        tail_ranks = head_predict(batch, model)
        head_ranks = tail_predict(batch, model)
        ranks = torch.cat([tail_ranks, head_ranks])
    elif predicion == "head":
        ranks = head_predict(batch, model)
    elif predicion == "tail":
        ranks = tail_predict(batch, model)

    return ranks.float()


def head_predict(batch, model):
    pos_triple = batch["positive_sample"]
    idx = pos_triple[:, 0]
    label = batch["head_label"]
    return calc_ranks(pos_triple, idx, label, "head-batch", model)


def tail_predict(batch, model):
    pos_triple = batch["positive_sample"]
    idx = pos_triple[:, 2]
    label = batch["tail_label"]
    return calc_ranks(pos_triple, idx, label, "tail-batch", model)


def calc_ranks(pos_triple, idx, label, mode, model):
    pred = model(pos_triple, mode=mode)
    b_range = torch.arange(pred.size()[0])
    target_pred = pred[b_range, idx]
    pred = torch.where(label.byte(), -torch.ones_like(pred) * 10000000, pred)
    pred[b_range, idx] = target_pred

    ranks = (
        1
        + torch.argsort(
            torch.argsort(pred, dim=1, descending=True), dim=1, descending=False
        )[b_range, idx]
    )
    return ranks


import torch
import torch.nn.functional as F
import torch.nn as nn
from IPython import embed

class Cross_Entropy_Loss(nn.Module):
    def __init__(self, args, model):
        super(Cross_Entropy_Loss, self).__init__()
        self.args = args
        self.model = model
        self.loss = torch.nn.BCELoss()

    def forward(self, pred, label):
        loss = self.loss(pred, label)
        return loss


import torch
import torch.nn as nn
import torch.nn.functional as F
from IPython import embed

class IterE_Loss(nn.Module):
    def __init__(self, args, model):
        super(IterE_Loss, self).__init__()
        self.args = args
        self.model = model
        #self.rule_p, self.rule_q = model.rule
        #self.confidence = model.conf
        
    def forward(self, pos_score, neg_score, subsampling_weight=None):
        if self.args.negative_adversarial_sampling:
            neg_score = (F.softmax(neg_score * self.args.adv_temp, dim=1).detach()
                        * F.logsigmoid(-neg_score)).sum(dim=1)  #shape:[bs]
        else:
            neg_score = F.logsigmoid(-neg_score).mean(dim = 1)

        pos_score = F.logsigmoid(pos_score).view(neg_score.shape[0]) #shape:[bs]

        if self.args.use_weight:
            positive_sample_loss = - (subsampling_weight * pos_score).sum()/subsampling_weight.sum()
            negative_sample_loss = - (subsampling_weight * neg_score).sum()/subsampling_weight.sum()
        else:
            positive_sample_loss = - pos_score.mean()
            negative_sample_loss = - neg_score.mean()

        loss = (positive_sample_loss + negative_sample_loss) / 2

        if True:
            #Use L3 regularization for ComplEx and DistMult
            regularization = self.args.regularization * (
                self.model.ent_emb.weight.norm(p = 3)**3 + \
                self.model.rel_emb.weight.norm(p = 3)**3
            )
            loss = loss + regularization
        return loss
    
    def normalize(self):
        regularization = self.args.regularization * (
                self.model.ent_emb.weight.norm(p = 3)**3 + \
                self.model.rel_emb.weight.norm(p = 3)**3
            )
        return regularization


import torch.nn as nn
import torch
from .model import Model
from IPython import embed


class ComplEx(Model):
    def __init__(self, args):
        """`Complex Embeddings for Simple Link Prediction`_ (ComplEx), which is a simple approach to matrix and tensor factorization for link prediction data that uses vectors with complex values and retains the mathematical definition of the dot product.

        Attributes:
            args: Model configuration parameters.
            epsilon: Calculate embedding_range.
            margin: Calculate embedding_range and loss.
            embedding_range: Uniform distribution range.
            ent_emb: Entity embedding, shape:[num_ent, emb_dim * 2].
            rel_emb: Relation_embedding, shape:[num_rel, emb_dim * 2].

        .. _Complex Embeddings for Simple Link Prediction: http://proceedings.mlr.press/v48/trouillon16.pdf
        """
        super(ComplEx, self).__init__(args)
        self.args = args
        self.ent_emb = None
        self.rel_emb = None

        self.init_emb()

    def init_emb(self):
        """Initialize the entity and relation embeddings in the form of a uniform distribution."""
        self.epsilon = 2.0
        self.margin = nn.Parameter(
            torch.Tensor([self.args.margin]), 
            requires_grad=False
        )
        self.embedding_range = nn.Parameter(
            torch.Tensor([(self.margin.item() + self.epsilon) / self.args.emb_dim]), 
            requires_grad=False
        )
        
        self.ent_emb = nn.Embedding(self.args.num_ent, self.args.emb_dim * 2)
        self.rel_emb = nn.Embedding(self.args.num_rel, self.args.emb_dim * 2)
        nn.init.uniform_(tensor=self.ent_emb.weight.data, a=-self.embedding_range.item(), b=self.embedding_range.item())
        nn.init.uniform_(tensor=self.rel_emb.weight.data, a=-self.embedding_range.item(), b=self.embedding_range.item())
        

    def score_func(self, head_emb, relation_emb, tail_emb, mode):
        """Calculating the score of triples.

        The formula for calculating the score is :math:`\operatorname{Re}\left(h^{\top} \operatorname{diag}(r) \overline{t}\right)`

        Args:
            head_emb: The head entity embedding.
            relation_emb: The relation embedding.
            tail_emb: The tail entity embedding.
            mode: Choose head-predict or tail-predict.

        Returns:
            score: The score of triples.
        """
        re_head, im_head = torch.chunk(head_emb, 2, dim=-1)
        re_relation, im_relation = torch.chunk(relation_emb, 2, dim=-1)
        re_tail, im_tail = torch.chunk(tail_emb, 2, dim=-1)

        return torch.sum(
            re_head * re_tail * re_relation
            + im_head * im_tail * re_relation
            + re_head * im_tail * im_relation
            - im_head * re_tail * im_relation,
            -1
        )
        

    def forward(self, triples, negs=None, mode='single'):
        """The functions used in the training phase

        Args:
            triples: The triples ids, as (h, r, t), shape:[batch_size, 3].
            negs: Negative samples, defaults to None.
            mode: Choose head-predict or tail-predict, Defaults to 'single'.

        Returns:
            score: The score of triples.
        """
        head_emb, relation_emb, tail_emb = self.tri2emb(triples, negs, mode)
        score = self.score_func(head_emb, relation_emb, tail_emb, mode)

        return score

    def get_score(self, batch, mode):
        """The functions used in the testing phase

        Args:
            batch: A batch of data.
            mode: Choose head-predict or tail-predict.

        Returns:
            score: The score of triples.
        """
        triples = batch["positive_sample"]
        head_emb, relation_emb, tail_emb = self.tri2emb(triples, mode=mode)
        score = self.score_func(head_emb, relation_emb, tail_emb, mode)
        return score


import torch
from torch import nn
import dgl
import dgl.function as fn
import torch.nn.functional as F
from neuralkg.model import ConvE

class CompGCN(nn.Module):

    """`Composition-based multi-relational graph convolutional networks`_ (CompGCN), 
        which jointly embeds both nodes and relations in a relational graph.

    Attributes:
        args: Model configuration parameters.
    
    .. _Composition-based multi-relational graph convolutional networks: 
        https://arxiv.org/pdf/1911.03082.pdf
    """

    def __init__(self, args):
        super(CompGCN, self).__init__()
        self.args      = args
        self.ent_emb   = None
        self.rel_emb   = None 
        self.GraphCov  = None  

        self.init_model()

    def init_model(self):
        """Initialize the CompGCN model and embeddings 

        Args:
            ent_emb: Entity embedding, shape:[num_ent, emb_dim].
            rel_emb: Relation_embedding, shape:[num_rel, emb_dim].
            GraphCov: The comp graph convolution layers.
            conv1: The convolution layer.
            fc: The full connection layer.
            bn0, bn1, bn2: The batch Normalization layer.
            inp_drop, hid_drop, feg_drop: The dropout layer.
        """
        #------------------------------CompGCN--------------------------------------------------------------------
        self.ent_emb = nn.Parameter(torch.Tensor(self.args.num_ent, self.args.emb_dim))
        self.rel_emb = nn.Parameter(torch.Tensor(self.args.num_rel, self.args.emb_dim))

        nn.init.xavier_normal_(self.ent_emb, gain=nn.init.calculate_gain('relu'))
        nn.init.xavier_normal_(self.rel_emb, gain=nn.init.calculate_gain('relu'))

        self.GraphCov = CompGCNCov(self.args.emb_dim, self.args.emb_dim * 2, torch.tanh, \
                                    bias = 'False', drop_rate = 0.1, opn = self.args.opn)
        
        self.bias = nn.Parameter(torch.zeros(self.args.num_ent))
        self.drop = nn.Dropout(0.3)
        #-----------------------------ConvE-----------------------------------------------------------------------
        self.emb_ent = torch.nn.Embedding(self.args.num_ent, self.args.emb_dim*2)
        self.inp_drop = torch.nn.Dropout(self.args.inp_drop)
        self.hid_drop = torch.nn.Dropout(self.args.hid_drop)
        self.feg_drop = torch.nn.Dropout2d(self.args.fet_drop)

        self.conv1 = torch.nn.Conv2d(1, 200, (7, 7), 1, 0, bias=False)
        self.bn0 = torch.nn.BatchNorm2d(1)
        self.bn1 = torch.nn.BatchNorm2d(200)
        self.bn2 = torch.nn.BatchNorm1d(200)
        self.register_parameter('b', torch.nn.Parameter(torch.zeros(self.args.num_ent)))
        self.fc = torch.nn.Linear(39200, self.args.out_dim)

    def forward(self, graph, relation, norm, triples):
        """The functions used in the training phase

        Args:
            graph: The knowledge graph recorded in dgl.graph()
            relation: The relation id sampled in triples
            norm: The edge norm in graph 
            triples: The triples ids, as (h, r, t), shape:[batch_size, 3].

        Returns:
            score: The score of triples.
        """
        head, rela = triples[:,0], triples[:, 1]
        x, r = self.ent_emb, self.rel_emb  # embedding of relations
        x, r = self.GraphCov(graph, x, r, relation, norm)
        x = self.drop(x)  # embeddings of entities [num_ent, dim]
        head_emb = torch.index_select(x, 0, head)  # filter out embeddings of subjects in this batch
        #head_in_emb = head_emb.view(-1, 1, 10, 20)

        rela_emb = torch.index_select(r, 0, rela)  # filter out embeddings of relations in this batch
        #rela_in_emb = rela_emb.view(-1, 1, 10, 20)

        if self.args.decoder_model.lower() == 'conve':
           # score = ConvE.score_func(self, head_in_emb, rela_in_emb, x)
           score = self.ConvE(head_emb, rela_emb, x)

        elif self.args.decoder_model.lower() == 'distmult':
            score = self.DistMult(head_emb, rela_emb)
        
        else:
            raise ValueError("^")

        return score

    def DistMult(self, head_emb, rela_emb):
        """Calculating the score of triples with DistMult model."""
        obj_emb = head_emb * rela_emb  # [batch_size, emb_dim]
        x = torch.mm(obj_emb, self.emb_ent.weight.transpose(1, 0))  # [batch_size, ent_num]
        x += self.bias.expand_as(x)
        score = torch.sigmoid(x)
        return score
    
    def concat(self, ent_embed, rel_embed):
        ent_embed = ent_embed.view(-1, 1, 200)
        rel_embed = rel_embed.view(-1, 1, 200)
        stack_input = torch.cat([ent_embed, rel_embed], 1)  # [batch_size, 2, embed_dim]
        stack_input = stack_input.reshape(-1, 1, 2 * 10, 20)  # reshape to 2D [batch, 1, 2*k_h, k_w]
        return stack_input

    def ConvE(self, sub_emb, rel_emb, all_ent):
        """Calculating the score of triples with ConvE model."""
        stack_input = self.concat(sub_emb, rel_emb)  # [batch_size, 1, 2*k_h, k_w]
        x = self.bn0(stack_input)
        x = self.conv1(x)  # [batch_size, num_filt, flat_sz_h, flat_sz_w]
        x = self.bn1(x)
        x = F.relu(x)
        x = self.feg_drop(x)
        x = x.view(x.shape[0], -1)  # [batch_size, flat_sz]
        x = self.fc(x)  # [batch_size, embed_dim]
        x = self.hid_drop(x)
        x = self.bn2(x)
        x = F.relu(x)
        x = torch.mm(x, all_ent.transpose(1, 0))  # [batch_size, ent_num]
        x += self.bias.expand_as(x)
        score = torch.sigmoid(x)
        return score

class CompGCNCov(nn.Module):
    """ The comp graph convolution layers, similar to https://github.com/malllabiisc/CompGCN"""
    def __init__(self, in_channels, out_channels, act=lambda x: x, bias=True, drop_rate=0., opn='corr'):
        super(CompGCNCov, self).__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.act = act  # activation function
        self.device = None
        self.rel = None
        self.opn = opn
        # relation-type specific parameter
        self.in_w = self.get_param([in_channels, out_channels])
        self.out_w = self.get_param([in_channels, out_channels])
        self.loop_w = self.get_param([in_channels, out_channels])
        self.w_rel = self.get_param([in_channels, out_channels])  # transform embedding of relations to next layer
        self.loop_rel = self.get_param([1, in_channels])  # self-loop embedding

        self.drop = nn.Dropout(drop_rate)
        self.bn = torch.nn.BatchNorm1d(out_channels)
        self.bias = nn.Parameter(torch.zeros(out_channels)) if bias else None
        self.rel_wt = None

    def get_param(self, shape):
        param = nn.Parameter(torch.Tensor(*shape))
        nn.init.xavier_normal_(param, gain=nn.init.calculate_gain('relu'))
        return param

    def message_func(self, edges: dgl.udf.EdgeBatch):
        edge_type = edges.data['type']  # [E, 1]
        edge_num = edge_type.shape[0]
        edge_data = self.comp(edges.src['h'], self.rel[edge_type])  # [E, in_channel]
        # msg = torch.bmm(edge_data.unsqueeze(1),
        #                 self.w[edge_dir.squeeze()]).squeeze()  # [E, 1, in_c] @ [E, in_c, out_c]
        # msg = torch.bmm(edge_data.unsqueeze(1),
        #                 self.w.index_select(0, edge_dir.squeeze())).squeeze()  # [E, 1, in_c] @ [E, in_c, out_c]
        # first half edges are all in-directions, last half edges are out-directions.
        msg = torch.cat([torch.matmul(edge_data[:edge_num // 2, :], self.in_w),
                         torch.matmul(edge_data[edge_num // 2:, :], self.out_w)])
        msg = msg * edges.data['norm'].reshape(-1, 1)  # [E, D] * [E, 1]
        return {'msg': msg}

    def reduce_func(self, nodes: dgl.udf.NodeBatch):
        return {'h': self.drop(nodes.data['h']) / 3}

    def comp(self, h, edge_data):
        def com_mult(a, b):
            r1, i1 = a.real, a.imag
            r2, i2 = b.real, b.imag
            real = r1 * r2 - i1 * i2
            imag = r1 * i2 + i1 * r2
            return torch.complex(real, imag)

        def conj(a):
            a.imag = -a.imag
            return a

        def ccorr(a, b):
            return torch.fft.irfft(com_mult(conj(torch.fft.rfft(a)), torch.fft.rfft(b)), a.shape[-1])

        if self.opn == 'mult':
            return h * edge_data
        elif self.opn == 'sub':
            return h - edge_data
        elif self.opn == 'corr':
            return ccorr(h, edge_data.expand_as(h))
        else:
            raise KeyError(f'composition operator {self.opn} not recognized.')

    def forward(self, g: dgl.graph, x, rel_repr, edge_type, edge_norm):
        self.device = x.device
        g = g.local_var()
        g.ndata['h'] = x
        g.edata['type'] = edge_type
        g.edata['norm'] = edge_norm
        if self.rel_wt is None:
            self.rel = rel_repr
        else:
            self.rel = torch.mm(self.rel_wt, rel_repr)  # [num_rel*2, num_base] @ [num_base, in_c]
        g.update_all(self.message_func, fn.sum(msg='msg', out='h'), self.reduce_func)
        x = g.ndata.pop('h') + torch.mm(self.comp(x, self.loop_rel), self.loop_w) / 3
        if self.bias is not None:
            x = x + self.bias
        x = self.bn(x)

        return self.act(x), torch.matmul(self.rel, self.w_rel)


import torch.nn as nn
import torch
from .model import Model
from IPython import embed


class RotatE(Model):
    """`RotatE: Knowledge Graph Embedding by Relational Rotation in Complex Space`_ (RotatE), which defines each relation as a rotation from the source entity to the target entity in the complex vector space.

    Attributes:
        args: Model configuration parameters.
        epsilon: Calculate embedding_range.
        margin: Calculate embedding_range and loss.
        embedding_range: Uniform distribution range.
        ent_emb: Entity embedding, shape:[num_ent, emb_dim * 2].
        rel_emb: Relation_embedding, shape:[num_rel, emb_dim].

    .. _RotatE: Knowledge Graph Embedding by Relational Rotation in Complex Space: https://openreview.net/forum?id=HkgEQnRqYQ
    """
    def __init__(self, args):
        super(RotatE, self).__init__(args)
        self.args = args
        self.ent_emb = None
        self.rel_emb = None

        self.init_emb()

    def init_emb(self):
        """Initialize the entity and relation embeddings in the form of a uniform distribution."""
        self.epsilon = 2.0
        self.margin = nn.Parameter(
            torch.Tensor([self.args.margin]), 
            requires_grad=False
        )
        self.embedding_range = nn.Parameter(
            torch.Tensor([(self.margin.item() + self.epsilon) / self.args.emb_dim]), 
            requires_grad=False
        )
        
        self.ent_emb = nn.Embedding(self.args.num_ent, self.args.emb_dim * 2)
        self.rel_emb = nn.Embedding(self.args.num_rel, self.args.emb_dim)
        nn.init.uniform_(tensor=self.ent_emb.weight.data, a=-self.embedding_range.item(), b=self.embedding_range.item())
        nn.init.uniform_(tensor=self.rel_emb.weight.data, a=-self.embedding_range.item(), b=self.embedding_range.item())
        

    def score_func(self, head_emb, relation_emb, tail_emb, mode):
        """Calculating the score of triples.

        The formula for calculating the score is :math:`\gamma - \|h \circ r - t\|`

        Args:
            head_emb: The head entity embedding.
            relation_emb: The relation embedding.
            tail_emb: The tail entity embedding.
            mode: Choose head-predict or tail-predict.

        Returns:
            score: The score of triples.
        """
        pi = 3.14159265358979323846
        re_head, im_head = torch.chunk(head_emb, 2, dim=-1)
        re_tail, im_tail = torch.chunk(tail_emb, 2, dim=-1)

        #Make phases of relations uniformly distributed in [-pi, pi]

        phase_relation = relation_emb/(self.embedding_range.item()/pi)

        re_relation = torch.cos(phase_relation)
        im_relation = torch.sin(phase_relation)

        if mode == 'head-batch':
            re_score = re_relation * re_tail + im_relation * im_tail
            im_score = re_relation * im_tail - im_relation * re_tail
            re_score = re_score - re_head
            im_score = im_score - im_head
        else:
            re_score = re_head * re_relation - im_head * im_relation
            im_score = re_head * im_relation + im_head * re_relation
            re_score = re_score - re_tail
            im_score = im_score - im_tail

        score = torch.stack([re_score, im_score], dim = 0)
        score = score.norm(dim = 0)
        score = self.margin.item() - score.sum(dim = -1)
        return score

    def forward(self, triples, negs=None, mode='single'):
        """The functions used in the training phase

        Args:
            triples: The triples ids, as (h, r, t), shape:[batch_size, 3].
            negs: Negative samples, defaults to None.
            mode: Choose head-predict or tail-predict, Defaults to 'single'.

        Returns:
            score: The score of triples.
        """
        head_emb, relation_emb, tail_emb = self.tri2emb(triples, negs, mode)
        score = self.score_func(head_emb, relation_emb, tail_emb, mode)

        return score

    def get_score(self, batch, mode):
        """The functions used in the testing phase

        Args:
            batch: A batch of data.
            mode: Choose head-predict or tail-predict.

        Returns:
            score: The score of triples.
        """
        triples = batch["positive_sample"]
        head_emb, relation_emb, tail_emb = self.tri2emb(triples, mode=mode)
        score = self.score_func(head_emb, relation_emb, tail_emb, mode)
        return score



import torch.nn as nn
import torch
from torch.autograd import Variable
from .model import Model

class BoxE(Model):
    
    """`A Box Embedding Model for Knowledge Base Completion`_ (BoxE), which represents the bump embedding as translations in the super rectangle space.

    Attributes:
        args: Model configuration parameters.
    
    .. _A Box Embedding Model for Knowledge Base Completion: https://arxiv.org/pdf/2007.06267.pdf
    """

    def __init__(self, args):
        super(BoxE, self).__init__(args)
        self.args = args
        self.arity = None
        self.order = None
        self.ent_emb = None
        self.rel_emb = None

        self.init_emb(args)

    def init_emb(self,args):

        """Initialize the entity and relation embeddings in the form of a uniform distribution.

        Args:
            arity: The maximum ary of the knowledge graph.
            epsilon: Caculate embedding_range.
            order: The distance order of score.
            margin: Caculate embedding_range and loss.
            embedding_range: Uniform distribution range.
            ent_emb: Entity embedding, shape:[num_ent, emb_dim * 2].
            rel_emb: Relation_embedding, shape:[num_rel, emb_dim * arity * 2].
        """
        
        self.arity = 2
        self.epsilon = 2.0
        self.order = self.args.dis_order
        self.margin = nn.Parameter(
            torch.Tensor([self.args.margin]),
            requires_grad=False
        )
        self.embedding_range = nn.Parameter(
            torch.Tensor([(self.margin.item() + self.epsilon) / self.args.emb_dim]),
            requires_grad=False
        )
        self.ent_emb = nn.Embedding(self.args.num_ent, self.args.emb_dim*2)
        nn.init.uniform_(tensor=self.ent_emb.weight.data[:, :self.args.emb_dim], a=-self.embedding_range.item(), b=self.embedding_range.item())
        nn.init.uniform_(tensor=self.ent_emb.weight.data[:, self.args.emb_dim:], a=-self.embedding_range.item(), b=self.embedding_range.item())
        size_factor = self.arity * 2
        self.rel_emb = nn.Embedding(self.args.num_rel, self.args.emb_dim * size_factor)

    def forward(self, triples, negs=None, mode='single'):

        """The functions used in the training phase

        Args:
            triples: The triples ids, as (h, r, t), shape:[batch_size, 3].
            negs: Negative samples, defaults to None.
            mode: Choose head-predict or tail-predict, Defaults to 'single'.

        Returns:
            score: The score of triples.
        """

        head_emb_raw, relation_emb, tail_emb_raw = self.tri2emb(triples, negs, mode)
        head_emb = head_emb_raw[:, :, :self.args.emb_dim] + tail_emb_raw[:, :, self.args.emb_dim:] 
        tail_emb = tail_emb_raw[:, :, :self.args.emb_dim] + head_emb_raw[:, :, self.args.emb_dim:]
        score = self.score_func(head_emb, relation_emb, tail_emb)
        return score
    
    def get_score(self, batch, mode):

        """The functions used in the testing phase

        Args:
            batch: A batch of data.
            mode: Choose head-predict or tail-predict.

        Returns:
            score: The score of triples.
        """

        triples = batch["positive_sample"]
        head_emb_raw, relation_emb, tail_emb_raw = self.tri2emb(triples, mode=mode)
        head_emb = head_emb_raw[:, :, :self.args.emb_dim] + tail_emb_raw[:, :, self.args.emb_dim:] 
        tail_emb = tail_emb_raw[:, :, :self.args.emb_dim] + head_emb_raw[:, :, self.args.emb_dim:]
        score = self.score_func(head_emb, relation_emb, tail_emb)
        return score
    
    def score_func(self, head_emb, relation_emb, tail_emb):

        """Calculate the score of the triple embedding.

        Args:
            head_emb: The embedding of head entity.
            relation_emb:The embedding of relation.
            tail_emb: The embedding of tail entity.

        Returns:
            score: The score of triples.
        """

        box_bas, box_del = torch.chunk(relation_emb, 2, dim = -1)
        box_sec = box_bas + 0.5 * box_del
        box_fir = box_bas - 0.5 * box_del
        box_low = torch.min(box_fir, box_sec)
        box_hig = torch.max(box_fir, box_sec)
        head_low, tail_low = torch.chunk(box_low, 2, dim = -1)
        head_hig, tail_hig = torch.chunk(box_hig, 2, dim = -1)
        head_score = self.calc_score(head_emb, head_low, head_hig, self.order)
        tail_score = self.calc_score(tail_emb, tail_low, tail_hig, self.order)
        score = self.margin.item() - (head_score + tail_score)
        return score
    
    def calc_score(self, ent_emb, box_low, box_hig, order = 2):
        
        """Calculate the norm of distance.

        Args:
            ent_emb: The embedding of entity.
            box_low: The lower boundaries of the super rectangle.
            box_hig: The upper boundaries of the super rectangle.
            order: The order of this distance.

        Returns:
            The norm of distance.
        """

        return torch.norm(self.dist(ent_emb, box_low, box_hig), p=order, dim=-1)
    
    def dist(self, ent_emb, lb, ub):

        """Calculate the distance.

        This function calculate the distance between the entity
        and the super rectangle. If the entity is in its target
        box, distance inversely correlates with box size, to
        maintain low distance inside large boxes and provide a
        gradient to keep points inside; if the entity is not in
        its target box, box size linearly correlates with distance,
        to penalize points outside larger boxes more severely.

        Args:
            ent_emb: The embedding of entity.
            lb: The lower boundaries of the super rectangle.
            ub: The upper boundaries of the super rectangle.

        Returns:
            The distance between entity and super rectangle.
        """

        c = (lb + ub) / 2
        w = ub - lb + 1
        k = 0.5 * (w - 1) * (w - 1 / w)
        return torch.where(torch.logical_and(torch.ge(ent_emb, lb), torch.le(ent_emb, ub)),
                        torch.abs(ent_emb - c) / w,
                        torch.abs(ent_emb - c) * w - k)
    






from numpy.random.mtrand import normal
import torch
import numpy as np
from torch.utils.data import Dataset
from collections import defaultdict as ddict
import random
from .DataPreprocess import *
from IPython import embed
import dgl 
import torch.nn.functional as F
import time
import queue

class UniSampler(BaseSampler):
    """Random negative sampling 
    Filtering out positive samples and selecting some samples randomly as negative samples.

    Attributes:
        cross_sampling_flag: The flag of cross sampling head and tail negative samples.
    """
    def __init__(self, args):
        super().__init__(args)
        self.cross_sampling_flag = 0

    def sampling(self, data):
        """Filtering out positive samples and selecting some samples randomly as negative samples.
        
        Args:
            data: The triples used to be sampled.

        Returns:
            batch_data: The training data.
        """
        batch_data = {}
        neg_ent_sample = []
        subsampling_weight = []
        self.cross_sampling_flag = 1 - self.cross_sampling_flag
        if self.cross_sampling_flag == 0:
            batch_data['mode'] = "head-batch"
            for h, r, t in data:
                neg_head = self.head_batch(h, r, t, self.args.num_neg)
                neg_ent_sample.append(neg_head)
                if self.args.use_weight:
                    weight = self.count[(h, r)] + self.count[(t, -r-1)]
                    subsampling_weight.append(weight)
        else:
            batch_data['mode'] = "tail-batch"
            for h, r, t in data:
                neg_tail = self.tail_batch(h, r, t, self.args.num_neg)
                neg_ent_sample.append(neg_tail)
                if self.args.use_weight:
                    weight = self.count[(h, r)] + self.count[(t, -r-1)]
                    subsampling_weight.append(weight)

        batch_data["positive_sample"] = torch.LongTensor(data)
        batch_data['negative_sample'] = torch.LongTensor(neg_ent_sample)
        if self.args.use_weight:
            batch_data["subsampling_weight"] = torch.sqrt(1/torch.tensor(subsampling_weight))
        return batch_data
    
    def uni_sampling(self, data):
        batch_data = {}
        neg_head_list = []
        neg_tail_list = []
        for h, r, t in data:
            neg_head = self.head_batch(h, r, t, self.args.num_neg)
            neg_head_list.append(neg_head)
            neg_tail = self.tail_batch(h, r, t, self.args.num_neg)
            neg_tail_list.append(neg_tail)

        batch_data["positive_sample"] = torch.LongTensor(data)
        batch_data['negative_head'] = torch.LongTensor(neg_head_list)
        batch_data['negative_tail'] = torch.LongTensor(neg_tail_list)
        return batch_data

    def get_sampling_keys(self):
        return ['positive_sample', 'negative_sample', 'mode']

class BernSampler(BaseSampler):
    """Using bernoulli distribution to select whether to replace the head entity or tail entity.
    
    Attributes:
        lef_mean: Record the mean of head entity
        rig_mean: Record the mean of tail entity
    """
    def __init__(self, args):
        super().__init__(args)
        self.lef_mean, self.rig_mean = self.calc_bern()
    def __normal_batch(self, h, r, t, neg_size):
        """Generate replace head/tail list according to Bernoulli distribution.
        
        Args:
            h: The head of triples.
            r: The relation of triples.
            t: The tail of triples.
            neg_size: The number of negative samples corresponding to each triple

        Returns:
             numpy.array: replace head list and replace tail list.
        """
        neg_size_h = 0
        neg_size_t = 0
        prob = self.rig_mean[r] / (self.rig_mean[r] + self.lef_mean[r]) if self.args.bern_flag else 0.5
        for i in range(neg_size):
            if random.random() < prob:
                neg_size_h += 1
            else:
                neg_size_t += 1

        neg_list_h = []
        neg_cur_size = 0
        while neg_cur_size < neg_size_h:
            neg_tmp_h = self.corrupt_head(t, r, num_max=(neg_size_h - neg_cur_size) * 2)
            neg_list_h.append(neg_tmp_h)
            neg_cur_size += len(neg_tmp_h)
        if neg_list_h != []:
            neg_list_h = np.concatenate(neg_list_h)

        neg_list_t = []
        neg_cur_size = 0
        while neg_cur_size < neg_size_t:
            neg_tmp_t = self.corrupt_tail(h, r, num_max=(neg_size_t - neg_cur_size) * 2)
            neg_list_t.append(neg_tmp_t)
            neg_cur_size += len(neg_tmp_t)
        if neg_list_t != []:
            neg_list_t = np.concatenate(neg_list_t)

        return np.hstack((neg_list_h[:neg_size_h], neg_list_t[:neg_size_t]))

    def sampling(self, data):
        """Using bernoulli distribution to select whether to replace the head entity or tail entity.
    
        Args:
            data: The triples used to be sampled.

        Returns:
            batch_data: The training data.
        """
        batch_data = {}
        neg_ent_sample = []
        batch_data['mode'] = 'bern'
        for h, r, t in data:
            neg_ent_sample = self.__normal_batch(h, r, t, self.args.num_neg)
        batch_data["positive_sample"] = torch.LongTensor(data)
        batch_data['negative_sample'] = torch.LongTensor(neg_ent_sample)
        return batch_data
    
    def calc_bern(self):
        """Calculating the lef_mean and rig_mean.
        
        Returns:
            lef_mean: Record the mean of head entity.
            rig_mean: Record the mean of tail entity.
        """
        h_of_r = ddict(set)
        t_of_r = ddict(set)
        freqRel = ddict(float)
        lef_mean = ddict(float)
        rig_mean = ddict(float)
        for h, r, t in self.train_triples:
            freqRel[r] += 1.0
            h_of_r[r].add(h)
            t_of_r[r].add(t)
        for r in h_of_r:
            lef_mean[r] = freqRel[r] / len(h_of_r[r])
            rig_mean[r] = freqRel[r] / len(t_of_r[r])
        return lef_mean, rig_mean

    @staticmethod
    def sampling_keys():
        return ['positive_sample', 'negative_sample', 'mode']

class AdvSampler(BaseSampler):
    """Self-adversarial negative sampling, in math:
    
    p\left(h_{j}^{\prime}, r, t_{j}^{\prime} \mid\left\{\left(h_{i}, r_{i}, t_{i}\right)\right\}\right)=\frac{\exp \alpha f_{r}\left(\mathbf{h}_{j}^{\prime}, \mathbf{t}_{j}^{\prime}\right)}{\sum_{i} \exp \alpha f_{r}\left(\mathbf{h}_{i}^{\prime}, \mathbf{t}_{i}^{\prime}\right)}
    
    Attributes:
        freq_hr: The count of (h, r) pairs.
        freq_tr: The count of (t, r) pairs.
    """
    def __init__(self, args):
        super().__init__(args)
        self.freq_hr, self.freq_tr = self.calc_freq()
    def sampling(self, pos_sample):
        """Self-adversarial negative sampling.
    
        Args:
            data: The triples used to be sampled.

        Returns:
            batch_data: The training data.
        """
        data = pos_sample.numpy().tolist()
        adv_sampling = []
        for h, r, t in data:
            weight = self.freq_hr[(h, r)] + self.freq_tr[(t, r)]
            adv_sampling.append(weight)
        adv_sampling = torch.tensor(adv_sampling, dtype=torch.float32).cuda()
        adv_sampling = torch.sqrt(1 / adv_sampling)
        return adv_sampling
    def calc_freq(self):
        """Calculating the freq_hr and freq_tr.
        
        Returns:
            freq_hr: The count of (h, r) pairs.
            freq_tr: The count of (t, r) pairs.
        """
        freq_hr, freq_tr = {}, {}
        for h, r, t in self.train_triples:
            if (h, r) not in freq_hr:
                freq_hr[(h, r)] = self.args.freq_init
            else:
                freq_hr[(h, r)] += 1
            if (t, r) not in freq_tr:
                freq_tr[(t, r)] = self.args.freq_init
            else:
                freq_tr[(t, r)] += 1
        return freq_hr, freq_tr

class AllSampler(RevSampler):
    """Merging triples which have same head and relation, all false tail entities are taken as negative samples.    
    """
    def __init__(self, args):
        super().__init__(args)
        # self.num_rel_without_rev = self.args.num_rel // 2
        
    def sampling(self, data):
        """Randomly sampling from the merged triples.
    
        Args:
            data: The triples used to be sampled.

        Returns:
            batch_data: The training data.
        """
        # sample_id = [] #确定triple里的relation是否是reverse的。reverse为1，不是为0
        batch_data = {}
        table = torch.zeros(len(data), self.args.num_ent)
        for id, (h, r, _) in enumerate(data):
            hr_sample = self.hr2t_train[(h, r)]
            table[id][hr_sample] = 1
            # if r > self.num_rel_without_rev:
            #     sample_id.append(1)
            # else:
            #     sample_id.append(0)
        batch_data["sample"] = torch.LongTensor(data)
        batch_data["label"] = table.float()
        # batch_data["sample_id"] = torch.LongTensor(sample_id)
        return batch_data

    def sampling_keys(self):
        return ["sample", "label"]
    
class CrossESampler(BaseSampler):
    # TODO:类名还需要商榷下
    def __init__(self, args):
        super().__init__(args)
        self.neg_weight = float(self.args.neg_weight / self.args.num_ent)
    def sampling(self, data):
        '''一个样本同时做head/tail prediction'''
        batch_data = {}
        hr_label = self.init_label(len(data))
        tr_label = self.init_label(len(data))
        for id, (h, r, t) in enumerate(data):
            hr_sample = self.hr2t_train[(h, r)]
            hr_label[id][hr_sample] = 1.0
            tr_sample = self.rt2h_train[(r, t)]
            tr_label[id][tr_sample] = 1.0
        batch_data["sample"] = torch.LongTensor(data)
        batch_data["hr_label"] = hr_label.float()
        batch_data["tr_label"] = tr_label.float()
        return batch_data

    def init_label(self, row):
        label = torch.rand(row, self.args.num_ent)
        label = (label > self.neg_weight).float()
        label -= 1.0
        return label

    def sampling_keys(self):
        return ["sample", "label"]

class ConvSampler(RevSampler):
    """Merging triples which have same head and relation, all false tail entities are taken as negative samples.      
    
    The triples which have same head and relation are treated as one triple.

    Attributes:
        label: Mask the false tail as negative samples.
        triples: The triples used to be sampled.
    """
    def __init__(self, args):
        self.label = None
        self.triples = None
        super().__init__(args)
        super().get_hr_trian()

    def sampling(self, pos_hr_t):
        """Randomly sampling from the merged triples.
    
        Args:
            pos_hr_t: The triples ((head,relation) pairs) used to be sampled.

        Returns:
            batch_data: The training data.
        """
        batch_data = {}

        self.label = torch.zeros(self.args.train_bs, self.args.num_ent)
        self.triples  = torch.LongTensor([hr for hr , _ in pos_hr_t])
        for id, hr_sample in enumerate([t for _ ,t in pos_hr_t]):
            self.label[id][hr_sample] = 1
    
        batch_data["sample"] = self.triples
        batch_data["label"] = self.label
        
        return batch_data

    def sampling_keys(self):
        return ["sample", "label"]

class XTransESampler(RevSampler):
    """Random negative sampling and recording neighbor entities.

    Attributes:
        triples: The triples used to be sampled.
        neg_sample: The negative samples.
        h_neighbor: The neighbor of sampled entites.
        h_mask: The tag of effecitve neighbor.
        max_neighbor: The maximum of the neighbor entities.
    """

    def __init__(self, args):
        super().__init__(args)
        super().get_h2rt_t2hr_from_train()
        self.triples    = None
        self.neg_sample = None
        self.h_neighbor = None
        self.h_mask     = None
        self.max_neighbor = 200

    def sampling(self, data):
        """Random negative sampling and recording neighbor entities.
    
        Args:
            data: The triples used to be sampled.

        Returns:
            batch_data: The training data.
        """
        batch_data = {}
        
        neg_ent_sample = []
        mask = np.zeros([self.args.train_bs, 20000], dtype=float)
        h_neighbor = np.zeros([self.args.train_bs, 20000, 2])
        
        for id, triples in enumerate(data):
            h,r,t = triples
            num_h_neighbor = len(self.h2rt_train[h]) 
            h_neighbor[id][0:num_h_neighbor] = np.array(self.h2rt_train[h])
            
            mask[id][0:num_h_neighbor] = np.ones([num_h_neighbor])
            
            neg_tail = self.tail_batch(h, r, t, self.args.num_neg)
            neg_ent_sample.append(neg_tail)

        self.triples    = data
        self.neg_sample = neg_ent_sample
        self.h_neighbor = h_neighbor[:, :self.max_neighbor]
        self.h_mask     = mask[:, :self.max_neighbor]

        batch_data["positive_sample"] = torch.LongTensor(self.triples)
        batch_data['negative_sample'] = torch.LongTensor(self.neg_sample)
        batch_data['neighbor']        = torch.LongTensor(self.h_neighbor)
        batch_data['mask']            = torch.LongTensor(self.h_mask)
        batch_data['mode']            = "tail-batch"
        return batch_data

    def get_sampling_keys(self):
        return ['positive_sample', 'negative_sample', 'neighbor', 'mask', 'mode']

class GraphSampler(RevSampler):
    """Graph based sampling in neural network.

    Attributes:
        entity: The entities of sampled triples. 
        relation: The relation of sampled triples.
        triples: The sampled triples.
        graph: The graph structured sampled triples by dgl.graph in DGL.
        norm: The edge norm in graph.
        label: Mask the false tail as negative samples.
    """
    def __init__(self, args):
        super().__init__(args)
        self.entity   = None
        self.relation = None
        self.triples  = None
        self.graph    = None
        self.norm     = None
        self.label    = None

    def sampling(self, pos_triples):
        """Graph based sampling in neural network.

        Args:
            pos_triples: The triples used to be sampled.

        Returns:
            batch_data: The training data.
        """
        batch_data = {}
        
        pos_triples = np.array(pos_triples)
        pos_triples, self.entity = self.sampling_positive(pos_triples)
        head_triples = self.sampling_negative('head', pos_triples, self.args.num_neg)
        tail_triples = self.sampling_negative('tail', pos_triples, self.args.num_neg)
        self.triples = np.concatenate((pos_triples,head_triples,tail_triples))
        batch_data['entity']  = self.entity
        batch_data['triples'] = self.triples
        
        self.label = torch.zeros((len(self.triples),1))
        self.label[0 : self.args.train_bs] = 1
        batch_data['label'] = self.label
        
        split_size = int(self.args.train_bs * 0.5) 
        graph_split_ids = np.random.choice(
            self.args.train_bs,
            size=split_size, 
            replace=False
        )
        head,rela,tail = pos_triples.transpose()
        head = torch.tensor(head[graph_split_ids], dtype=torch.long).contiguous()
        rela = torch.tensor(rela[graph_split_ids], dtype=torch.long).contiguous()
        tail = torch.tensor(tail[graph_split_ids], dtype=torch.long).contiguous()
        self.graph, self.relation, self.norm = self.build_graph(len(self.entity), (head,rela,tail), -1)
        batch_data['graph']    = self.graph
        batch_data['relation'] = self.relation
        batch_data['norm']     = self.norm

        return batch_data

    def get_sampling_keys(self):
        return ['graph','triples','label','entity','relation','norm']

    def sampling_negative(self, mode, pos_triples, num_neg):
        """Random negative sampling without filtering

        Args:
            mode: The mode of negtive sampling.
            pos_triples: The positive triples.
            num_neg: The number of negative samples corresponding to each triple.

        Results:
            neg_samples: The negative triples.
        """
        neg_random = np.random.choice(
            len(self.entity), 
            size = num_neg * len(pos_triples)
        )
        neg_samples = np.tile(pos_triples, (num_neg, 1))
        if mode == 'head':
            neg_samples[:,0] = neg_random
        elif mode == 'tail':
            neg_samples[:,2] = neg_random
        return neg_samples

    def build_graph(self, num_ent, triples, power):
        """Using sampled triples to build a graph by dgl.graph in DGL.

        Args:
            num_ent: The number of entities.
            triples: The positive sampled triples.
            power: The power index for normalization.

        Returns:
            rela: The relation of sampled triples.
            graph: The graph structured sampled triples by dgl.graph in DGL.
            edge_norm: The edge norm in graph.
        """
        head, rela, tail = triples[0], triples[1], triples[2]
        graph = dgl.graph(([], []))
        graph.add_nodes(num_ent)
        graph.add_edges(head, tail)
        node_norm = self.comp_deg_norm(graph, power)
        edge_norm = self.node_norm_to_edge_norm(graph,node_norm)
        rela = torch.tensor(rela)
        return graph, rela, edge_norm

    def comp_deg_norm(self, graph, power=-1):
        """Calculating the normalization node weight.

        Args:
            graph: The graph structured sampled triples by dgl.graph in DGL.
            power: The power index for normalization.

        Returns:
            tensor: The node weight of normalization.
        """
        graph = graph.local_var()
        in_deg = graph.in_degrees(range(graph.number_of_nodes())).float().numpy()
        norm = in_deg.__pow__(power)
        norm[np.isinf(norm)] = 0
        return torch.from_numpy(norm)

    def node_norm_to_edge_norm(slef, graph, node_norm):
        """Calculating the normalization edge weight.

        Args:
            graph: The graph structured sampled triples by dgl.graph in DGL.
            node_norm: The node weight of normalization.

        Returns:
            tensor: The edge weight of normalization.
        """
        graph = graph.local_var()
        # convert to edge norm
        graph.ndata['norm'] = node_norm.view(-1,1)
        graph.apply_edges(lambda edges : {'norm' : edges.dst['norm']})
        return graph.edata['norm']

    def sampling_positive(self,positive_triples):
        """Regenerate positive sampling.

        Args:
            positive_triples: The positive sampled triples.

        Results:
            The regenerate triples and entities filter invisible entities.
        """

        edges = np.random.choice(
            np.arange(len(positive_triples)),
            size = self.args.train_bs,
            replace=False
        )
        edges = positive_triples[edges]
        head, rela, tail = np.array(edges).transpose()
        entity, index = np.unique((head, tail), return_inverse=True) 
        head, tail = np.reshape(index, (2, -1))

        return np.stack((head,rela,tail)).transpose(), \
                torch.from_numpy(entity).view(-1,1).long()

class KBATSampler(BaseSampler):
    """Graph based n_hop neighbours in neural network.

    Attributes:
        n_hop: The graph of n_hop neighbours.
        graph: The adjacency graph.
        neighbours: The neighbours of sampled triples.
        adj_matrix:The triples of sampled.
        triples: The sampled triples.
        triples_GAT_pos: Positive triples.
        triples_GAT_neg: Negative triples.
        triples_Con: All triples including positive triples and negative triples. 
        label: Mask the false tail as negative samples.
    """
    def __init__(self, args): 
        super().__init__(args)
        self.n_hop           = None
        self.graph           = None
        self.neighbours      = None
        self.adj_matrix      = None
        self.entity          = None
        self.triples_GAT_pos = None
        self.triples_GAT_neg = None
        self.triples_Con     = None
        self.label           = None

        self.get_neighbors()

    def sampling(self, pos_triples):
        """Graph based n_hop neighbours in neural network.

        Args:
            pos_triples: The triples used to be sampled.

        Returns:
            batch_data: The training data.
        """
        batch_data = {}
        #--------------------KBAT-Sampler------------------------------------------
        self.entity = self.get_unique_entity(pos_triples)
        head_triples = self.sam_negative('head', pos_triples, self.args.num_neg)
        tail_triples = self.sam_negative('tail', pos_triples, self.args.num_neg)
        self.triples_GAT_neg = torch.tensor(np.concatenate((head_triples, tail_triples)))
        batch_data['triples_GAT_pos'] = torch.tensor(pos_triples)
        batch_data['triples_GAT_neg'] = self.triples_GAT_neg

        head, rela, tail = torch.tensor(self.train_triples).t()
        self.adj_matrix  = (torch.stack((tail, head)), rela)
        batch_data['adj_matrix'] = self.adj_matrix

        self.n_hop = self.get_batch_nhop_neighbors_all()
        batch_data['n_hop'] = self.n_hop
        #--------------------ConvKB-Sampler------------------------------------------
        head_triples = self.sampling_negative('head', pos_triples, self.args.num_neg)
        tail_triples = self.sampling_negative('tail', pos_triples, self.args.num_neg)
        self.triples_Con = np.concatenate((pos_triples, head_triples, tail_triples))
        self.label = -torch.ones((len(self.triples_Con),1))
        self.label[0 : self.args.train_bs] = 1
        batch_data['triples_Con'] = self.triples_Con
        batch_data['label'] = self.label

        return batch_data

    def get_sampling_keys(self):
        return ['adj_matrix', 'n_hop', 'triples_GAT_pos', 
        'triples_GAT_neg', 'triples_Con' , 'label']

    def bfs(self, graph, source, nbd_size=2):
        """Using depth first search algorithm to generate n_hop neighbor graph.
        
        Args:
            graph: The adjacency graph.
            source: Head node.
            nbd_size: The number of hops.

        Returns:
            neighbors: N_hop neighbor graph.
        """
        visit = {}
        distance = {}
        parent = {}
        distance_lengths = {}

        visit[source] = 1
        distance[source] = 0
        parent[source] = (-1, -1)

        q = queue.Queue()
        q.put((source, -1))

        while(not q.empty()):
            top = q.get()
            if top[0] in graph.keys():
                for target in graph[top[0]].keys():
                    if(target in visit.keys()):
                        continue
                    else:
                        q.put((target, graph[top[0]][target]))

                        distance[target] = distance[top[0]] + 1

                        visit[target] = 1
                        if distance[target] > 2:
                            continue
                        parent[target] = (top[0], graph[top[0]][target]) # 记录父亲节点id和关系id

                        if distance[target] not in distance_lengths.keys():
                            distance_lengths[distance[target]] = 1

        neighbors = {}
        for target in visit.keys():
            if(distance[target] != nbd_size):
                continue
            edges = [-1, parent[target][1]]
            relations = []
            entities = [target]
            temp = target
            while(parent[temp] != (-1, -1)):
                relations.append(parent[temp][1])
                entities.append(parent[temp][0])
                temp = parent[temp][0]

            if(distance[target] in neighbors.keys()):
                neighbors[distance[target]].append(
                    (tuple(relations), tuple(entities[:-1]))) #删除已知的source 记录前两跳实体及关系
            else:
                neighbors[distance[target]] = [
                    (tuple(relations), tuple(entities[:-1]))]

        return neighbors

    def get_neighbors(self, nbd_size=2):
        """Getting the relation and entity of the source in the n_hop neighborhood.
        
        Args:
            nbd_size: The number of hops.

        Returns:
            self.neighbours: Record the relation and entity of the source in the n_hop neighborhood.
        """
        self.graph = {}

        for triple in self.train_triples:
            head = triple[0]
            rela = triple[1]
            tail = triple[2]

            if(head not in self.graph.keys()):
                self.graph[head] = {}
                self.graph[head][tail] = rela
            else:
                self.graph[head][tail] = rela

        neighbors = {}
        '''
        import pickle
        print("Opening node_neighbors pickle object")
        file = self.args.data_path + "/2hop.pickle"
        with open(file, 'rb') as handle:
            self.neighbours = pickle.load(handle)  
        return
        '''
        start_time = time.time()
        print("Start Graph BFS")
        for head in self.graph.keys():
            temp_neighbors = self.bfs(self.graph, head, nbd_size)
            for distance in temp_neighbors.keys():
                if(head in neighbors.keys()):
                    if(distance in neighbors[head].keys()):
                        neighbors[head][distance].append(
                            temp_neighbors[distance])
                    else:
                        neighbors[head][distance] = temp_neighbors[distance]
                else:
                    neighbors[head] = {}
                    neighbors[head][distance] = temp_neighbors[distance]

        print("Finish BFS, time taken ", time.time() - start_time)
        self.neighbours = neighbors

    def get_unique_entity(self, triples):
        """Getting the set of entity.
        
        Args:
            triples: The sampled triples.

        Returns:
            numpy.array: The set of entity
        """
        train_triples = np.array(triples)
        train_entities = np.concatenate((train_triples[:,0], train_triples[:,2]))
        return np.unique(train_entities)

    def get_batch_nhop_neighbors_all(self, nbd_size=2):
        """Getting n_hop neighbors of all entities in batch.
        
        Args:
            nbd_size: The number of hops.

        Returns:
            The set of n_hop neighbors.
        """
        batch_source_triples = []
        
        for source in self.entity:
            if source in self.neighbours.keys():
                nhop_list = self.neighbours[source][nbd_size]
                for i, tup in enumerate(nhop_list):
                    if(i >= 2): 
                        break
                    batch_source_triples.append([source, 
                                                tup[0][-1], 
                                                tup[0][0],
                                                tup[1][0]])

        n_hop =  np.array(batch_source_triples).astype(np.int32)
        
        return torch.autograd.Variable(torch.LongTensor(n_hop))

    def sampling_negative(self, mode, pos_triples, num_neg):
        """Random negative sampling.

        Args:
            mode: The mode of negtive sampling.
            pos_triples: The positive triples.
            num_neg: The number of negative samples corresponding to each triple.

        Results:
            neg_samples: The negative triples.
        """
        neg_samples = np.tile(pos_triples, (num_neg, 1))
        if mode == 'head':
            neg_head = []
            for h, r, t in pos_triples:
                neg_head.append(self.head_batch(h, r, t, num_neg))
            neg_samples[:,0] = torch.tensor(neg_head).t().reshape(-1)
        elif mode == 'tail':
            neg_tail = []
            for h, r, t in pos_triples:
                neg_tail.append(self.tail_batch(h, r, t, num_neg))
            neg_samples[:,2] = torch.tensor(neg_tail).t().reshape(-1)
        return neg_samples

    def sam_negative(self, mode, pos_triples, num_neg):
        """Random negative sampling without filter.

        Args:
            mode: The mode of negtive sampling.
            pos_triples: The positive triples.
            num_neg: The number of negative samples corresponding to each triple.

        Results:
            neg_samples: The negative triples.
        """ 
        neg_random = np.random.choice(
            len(self.entity), 
            size = num_neg * len(pos_triples)
        )
        neg_samples = np.tile(pos_triples, (num_neg, 1))
        if mode == 'head':
            neg_samples[:,0] = neg_random
        elif mode == 'tail':
            neg_samples[:,2] = neg_random
        return neg_samples

class CompGCNSampler(GraphSampler):
    """Graph based sampling in neural network.

    Attributes:
        relation: The relation of sampled triples.
        triples: The sampled triples.
        graph: The graph structured sampled triples by dgl.graph in DGL.
        norm: The edge norm in graph.
        label: Mask the false tail as negative samples.
    """
    def __init__(self, args):
        super().__init__(args)
        self.relation = None
        self.triples  = None
        self.graph    = None
        self.norm     = None
        self.label    = None
        
        super().get_hr_trian()
        
        self.graph, self.relation, self.norm = \
            self.build_graph(self.args.num_ent, np.array(self.t_triples).transpose(), -0.5)

    def sampling(self, pos_hr_t):
        """Graph based n_hop neighbours in neural network.

        Args:
            pos_hr_t: The triples(hr, t) used to be sampled.

        Returns:
            batch_data: The training data.
        """
        batch_data = {}
        
        self.label = torch.zeros(self.args.train_bs, self.args.num_ent)
        self.triples  = torch.LongTensor([hr for hr , _ in pos_hr_t])
        for id, hr_sample in enumerate([t for _ ,t in pos_hr_t]):
            self.label[id][hr_sample] = 1

        batch_data['sample']   = self.triples
        batch_data['label']    = self.label
        batch_data['graph']    = self.graph
        batch_data['relation'] = self.relation
        batch_data['norm']     = self.norm

        return batch_data

    def get_sampling_keys(self):
        return ['sample','label','graph','relation','norm']

    def node_norm_to_edge_norm(self, graph, node_norm):
        """Calculating the normalization edge weight.

        Args:
            graph: The graph structured sampled triples by dgl.graph in DGL.
            node_norm: The node weight of normalization.

        Returns:
            norm: The edge weight of normalization.
        """
        graph.ndata['norm'] = node_norm
        graph.apply_edges(lambda edges: {'norm': edges.dst['norm'] * edges.src['norm']})
        norm = graph.edata.pop('norm').squeeze()
        return norm

class TestSampler(object):
    """Sampling triples and recording positive triples for testing.

    Attributes:
        sampler: The function of training sampler.
        hr2t_all: Record the tail corresponding to the same head and relation.
        rt2h_all: Record the head corresponding to the same tail and relation.
        num_ent: The count of entities.
    """
    def __init__(self, sampler):
        self.sampler = sampler
        self.hr2t_all = ddict(set)
        self.rt2h_all = ddict(set)
        self.get_hr2t_rt2h_from_all()
        self.num_ent = sampler.args.num_ent

    def get_hr2t_rt2h_from_all(self):
        """Get the set of hr2t and rt2h from all datasets(train, valid, and test), the data type is tensor.

        Update:
            self.hr2t_all: The set of hr2t.
            self.rt2h_all: The set of rt2h.
        """
        self.all_true_triples = self.sampler.get_all_true_triples()
        for h, r, t in self.all_true_triples:
            self.hr2t_all[(h, r)].add(t)
            self.rt2h_all[(r, t)].add(h)
        for h, r in self.hr2t_all:
            self.hr2t_all[(h, r)] = torch.tensor(list(self.hr2t_all[(h, r)]))
        for r, t in self.rt2h_all:
            self.rt2h_all[(r, t)] = torch.tensor(list(self.rt2h_all[(r, t)]))

    def sampling(self, data):
        """Sampling triples and recording positive triples for testing.

        Args:
            data: The triples used to be sampled.

        Returns:
            batch_data: The data used to be evaluated.
        """
        batch_data = {}
        head_label = torch.zeros(len(data), self.num_ent)
        tail_label = torch.zeros(len(data), self.num_ent)
        for idx, triple in enumerate(data):
            head, rel, tail = triple
            head_label[idx][self.rt2h_all[(rel, tail)]] = 1.0
            tail_label[idx][self.hr2t_all[(head, rel)]] = 1.0
        batch_data["positive_sample"] = torch.tensor(data)
        batch_data["head_label"] = head_label
        batch_data["tail_label"] = tail_label
        return batch_data

    def get_sampling_keys(self):
        return ["positive_sample", "head_label", "tail_label"]

class GraphTestSampler(object):
    """Sampling graph for testing.

    Attributes:
        sampler: The function of training sampler.
        hr2t_all: Record the tail corresponding to the same head and relation.
        rt2h_all: Record the head corresponding to the same tail and relation.
        num_ent: The count of entities.
        triples: The training triples.
    """
    def __init__(self, sampler):
        self.sampler = sampler
        self.hr2t_all = ddict(set)
        self.rt2h_all = ddict(set)
        self.get_hr2t_rt2h_from_all()
        self.num_ent = sampler.args.num_ent
        self.triples = sampler.train_triples

    def get_hr2t_rt2h_from_all(self):
        """Get the set of hr2t and rt2h from all datasets(train, valid, and test), the data type is tensor.

        Update:
            self.hr2t_all: The set of hr2t.
            self.rt2h_all: The set of rt2h.
        """
        self.all_true_triples = self.sampler.get_all_true_triples()
        for h, r, t in self.all_true_triples:
            self.hr2t_all[(h, r)].add(t)
            self.rt2h_all[(r, t)].add(h)
        for h, r in self.hr2t_all:
            self.hr2t_all[(h, r)] = torch.tensor(list(self.hr2t_all[(h, r)]))
        for r, t in self.rt2h_all:
            self.rt2h_all[(r, t)] = torch.tensor(list(self.rt2h_all[(r, t)]))

    def sampling(self, data):
        """Sampling graph for testing.

        Args:
            data: The triples used to be sampled.

        Returns:
            batch_data: The data used to be evaluated.
        """
        batch_data = {}
        head_label = torch.zeros(len(data), self.num_ent)
        tail_label = torch.zeros(len(data), self.num_ent)
        for idx, triple in enumerate(data):
            # from IPython import embed;embed();exit()
            head, rel, tail = triple
            head_label[idx][self.rt2h_all[(rel, tail)]] = 1.0
            tail_label[idx][self.hr2t_all[(head, rel)]] = 1.0
        batch_data["positive_sample"] = torch.tensor(data)
        batch_data["head_label"] = head_label
        batch_data["tail_label"] = tail_label
        
        head, rela, tail = np.array(self.triples).transpose()
        graph, rela, norm = self.sampler.build_graph(self.num_ent, (head, rela, tail), -1)
        batch_data["graph"]  = graph
        batch_data["rela"]   = rela
        batch_data["norm"]   = norm
        batch_data["entity"] = torch.arange(0, self.num_ent, dtype=torch.long).view(-1,1)
        
        return batch_data

    def get_sampling_keys(self):
        return ["positive_sample", "head_label", "tail_label",\
             "graph", "rela", "norm", "entity"]

class CompGCNTestSampler(object):  #TODO: 争取和GraphTestSampler兼容
    """Sampling graph for testing.

    Attributes:
        sampler: The function of training sampler.
        hr2t_all: Record the tail corresponding to the same head and relation.
        rt2h_all: Record the head corresponding to the same tail and relation.
        num_ent: The count of entities.
        triples: The training triples.
    """
    def __init__(self, sampler):
        self.sampler = sampler
        self.hr2t_all = ddict(set)
        self.rt2h_all = ddict(set)
        self.get_hr2t_rt2h_from_all()
        self.num_ent = sampler.args.num_ent
        self.triples = sampler.t_triples

    def get_hr2t_rt2h_from_all(self):
        """Get the set of hr2t and rt2h from all datasets(train, valid, and test), the data type is tensor.

        Update:
            self.hr2t_all: The set of hr2t.
            self.rt2h_all: The set of rt2h.
        """
        self.all_true_triples = self.sampler.get_all_true_triples()
        for h, r, t in self.all_true_triples:
            self.hr2t_all[(h, r)].add(t)
            self.rt2h_all[(r, t)].add(h)
        for h, r in self.hr2t_all:
            self.hr2t_all[(h, r)] = torch.tensor(list(self.hr2t_all[(h, r)]))
        for r, t in self.rt2h_all:
            self.rt2h_all[(r, t)] = torch.tensor(list(self.rt2h_all[(r, t)]))

    def sampling(self, data):
        """Sampling graph for testing.

        Args:
            data: The triples used to be sampled.

        Returns:
            batch_data: The data used to be evaluated.
        """
        batch_data = {}
        
        head_label = torch.zeros(len(data), self.num_ent)
        tail_label = torch.zeros(len(data), self.num_ent)
        
        for idx, triple in enumerate(data):
            # from IPython import embed;embed();exit()
            head, rel, tail = triple
            head_label[idx][self.rt2h_all[(rel, tail)]] = 1.0
            tail_label[idx][self.hr2t_all[(head, rel)]] = 1.0
        batch_data["positive_sample"] = torch.tensor(data)
        batch_data["head_label"] = head_label
        batch_data["tail_label"] = tail_label
        
        graph, relation, norm = \
            self.sampler.build_graph(self.num_ent, np.array(self.triples).transpose(), -0.5)
    
        batch_data["graph"]  = graph
        batch_data["rela"]   = relation
        batch_data["norm"]   = norm
        batch_data["entity"] = torch.arange(0, self.num_ent, dtype=torch.long).view(-1,1)
        
        return batch_data

    def get_sampling_keys(self):
        return ["positive_sample", "head_label", "tail_label",\
             "graph", "rela", "norm", "entity"]

class XTransETestSampler(TestSampler):
    """Sampling triples and recording positive triples for testing.

    Attributes:
        sampler: The function of training sampler.
        hr2t_all: Record the tail corresponding to the same head and relation.
        rt2h_all: Record the head corresponding to the same tail and relation.
    """
    def __init__(self, sampler):
        self.sampler = sampler
        self.hr2t_all = ddict(set)
        self.rt2h_all = ddict(set)
        super().get_hr2t_rt2h_from_all()
    
    def sampling(self, data):
        """Sampling triples and recording positive triples for testing.

        Args:
            data: The triples used to be sampled.

        Returns:
            batch_data: The data used to be evaluated.
        """
        batch_data = {}
        tail_label = torch.zeros(len(data), self.sampler.args.num_ent)
        mask = np.zeros([self.sampler.args.train_bs, 20000], dtype=float)    
        h_neighbor = np.zeros([self.sampler.args.train_bs, 20000, 2])   #分别记录关系和尾实体
        
        for idx, triple in enumerate(data):
            head, rel, tail = triple
            num_h_neighbor = len(self.sampler.h2rt_train[head])
            h_neighbor[id][0:num_h_neighbor] = np.asarray(self.sampler.h2rt_train[head])
            mask[id][0:num_h_neighbor] = np.ones([num_h_neighbor])
            tail_label[idx][self.sampler.hr2t_all[(head, rel)]] = 1.0
        
        batch_data["positive_sample"] = torch.tensor(data)
        batch_data["tail_label"] = tail_label
        batch_data['neighbor'] = h_neighbor[:,:self.sampler.max_neighbor]
        batch_data['mask'] = mask[:,:self.sampler.max_neighbor]

        return batch_data

    def get_sampling_keys(self):
        return ["positive_sample", "tail_label", "neighbor", "mask"]
    
'''继承torch.Dataset'''
class KGDataset(Dataset):

    def __init__(self, triples):
        self.triples = triples

    def __len__(self):
        return len(self.triples)

    def __getitem__(self, idx):
        return self.triples[idx]


import argparse
import pytorch_lightning as pl
import torch
from collections import defaultdict as ddict

from neuralkg import loss
import numpy as np

OPTIMIZER = "AdamW"
LR = 5e-5
LOSS = "cross_entropy"
ONE_CYCLE_TOTAL_STEPS = 100

class Config(dict):
    def __getattr__(self, name):
        return self.get(name)

    def __setattr__(self, name, val):
        self[name] = val


class BaseLitModel(pl.LightningModule):
    """
    Generic PyTorch-Lightning class that must be initialized with a PyTorch module.
    """

    def __init__(self, model, args: argparse.Namespace = None):
        super().__init__()
        self.model = model
        self.args = args
        optim_name = args.optim_name
        self.optimizer_class = getattr(torch.optim, optim_name)
        loss_name = args.loss_name
        self.loss_class = getattr(loss, loss_name)
        self.loss = self.loss_class(args, model)


    @staticmethod
    def add_to_argparse(parser):
        parser.add_argument("--lr", type=float, default=LR)
        parser.add_argument("--weight_decay", type=float, default=0.01)
        return parser

    def configure_optimizers(self):
        optimizer = self.optimizer_class(self.parameters(), lr=self.lr)
        if self.one_cycle_max_lr is None:
            return optimizer
        scheduler = torch.optim.lr_scheduler.OneCycleLR(optimizer=optimizer, max_lr=self.one_cycle_max_lr, total_steps=self.one_cycle_total_steps)
        return {"optimizer": optimizer, "lr_scheduler": scheduler, "monitor": "val_loss"}

    def forward(self, x):
        return self.model(x)

    def training_step(self, batch, batch_idx):  # pylint: disable=unused-argument
        x, y = batch
        logits = self(x)
        loss = self.loss_fn(logits, y)
        self.log("train_loss", loss)
        self.train_acc(logits, y)
        self.log("train_acc", self.train_acc, on_step=False, on_epoch=True)
        return loss

    def validation_step(self, batch, batch_idx):  # pylint: disable=unused-argument
        x, y = batch
        logits = self(x)
        loss = self.loss_fn(logits, y)
        self.log("val_loss", loss, prog_bar=True)
        self.val_acc(logits, y)
        self.log("val_acc", self.val_acc, on_step=False, on_epoch=True, prog_bar=True)

    def test_step(self, batch, batch_idx):  # pylint: disable=unused-argument
        x, y = batch
        logits = self(x)
        self.test_acc(logits, y)
        self.log("test_acc", self.test_acc, on_step=False, on_epoch=True)
    
    def collect_results(self, results, mode):
        """Summarize the results of each batch and calculate the final result of the epoch

        Args:
            results ([type]): The results of each batch
            mode ([type]): Eval or Test

        Returns:
            dict: The final result of the epoch
        """
        outputs = ddict(float)
        count = np.array([o["count"] for o in results]).sum()
        for metric in list(results[0].keys())[1:]:
            final_metric = "|".join([mode, metric])
            outputs[final_metric] = np.around(np.array([o[metric] for o in results]).sum() / count, decimals=3).item()
        return outputs



import random
import numpy as np
import torch
from torch.utils.data import Dataset, DataLoader
import os
from collections import defaultdict as ddict
from IPython import embed


class RuleDataset(Dataset):
    def __init__(self, args):
        self.args = args
        self.rule_p, self.rule_q, self.rule_r, self.confidences, self.tripleNum = [], [], [], [], []
        with open(os.path.join(args.data_path, 'groudings.txt')) as f:
            for line in f.readlines():
                token = line.strip().split('\t')
                for i in range(len(token)):
                    token[i] = token[i].strip('(').strip(')')

                iUnseenPos = int(token[0])
                self.tripleNum.append(iUnseenPos)

                iFstHead = int(token[1])
                iFstTail = int(token[3])
                iFstRelation = int(token[2])
                self.rule_p.append([iFstHead, iFstRelation, iFstTail])

                iSndHead = int(token[4])
                iSndTail = int(token[6])
                iSndRelation = int(token[5])
                self.rule_q.append([iSndHead, iSndRelation, iSndTail])

                if len(token) == 8:
                    confidence = float(token[7])
                    self.rule_r.append([0, 0, 0])

                else:
                    confidence = float(token[10])
                    iTrdHead = int(token[7])
                    iTrdTail = int(token[9])
                    iTrdRelation = int(token[8])
                    self.rule_r.append([iTrdHead, iTrdRelation, iTrdTail])
                self.confidences.append(confidence)
            self.len = len(self.confidences)
            self.rule_p = torch.tensor(self.rule_p).to(self.args.gpu)
            self.rule_q = torch.tensor(self.rule_q).to(self.args.gpu)
            self.rule_r = torch.tensor(self.rule_r).to(self.args.gpu)
            self.confidences = torch.tensor(self.confidences).to(self.args.gpu)
            self.tripleNum = torch.tensor(self.tripleNum).to(self.args.gpu)

    def __len__(self):
        return self.len

    def __getitem__(self, idx):
        return (self.rule_p[idx], self.rule_q[idx], self.rule_r[idx]), self.confidences[idx], self.tripleNum[idx]


class RuleDataLoader(DataLoader):
    def __init__(self, args):
        dataset = RuleDataset(args)
        super(RuleDataLoader, self).__init__(
            dataset=dataset,
            batch_size=int(dataset.__len__()/args.num_batches),
            shuffle=args.shuffle)


import torch.nn as nn
import torch
import torch.nn.functional as F
import math
from .model import Model
from IPython import embed


class CrossE(Model):
    """`Interaction Embeddings for Prediction and Explanation in Knowledge Graphs`_ (CrossE), which simulates crossover interactions(bi-directional effects between entities and relations)
    to select related information when predicting a new triple

    Attributes:
        args: Model configuration parameters.
        ent_emb: Entity embedding, shape:[num_ent, emb_dim].
        rel_emb: Relation embedding, shape:[num_rel, emb_dim].
        rel_reverse_emb: Reverse Relation embedding, shape:[num_rel, emb_dim].
        h_weighted_vector: Interaction matrix for head entities and relations, shape:[num_rel, emb_dim]
        t_weighted_vector: Interaction matrix for tail entities and relations, shape:[num_rel, emb_dim]
        hr_bias: Bias for head entity and relation
        tr_bias: Bias for tail entity and relation

    .. _Interaction Embeddings for Prediction and Explanation in Knowledge Graphs: https://arxiv.org/abs/1903.04750
    """
    def __init__(self, args):
        super(CrossE, self).__init__(args)
        self.args = args
        self.ent_emb = None
        self.rel_emb = None
        self.init_emb()

    def init_emb(self):

        self.dropout = nn.Dropout(self.args.dropout)
        self.ent_emb = nn.Embedding(self.args.num_ent, self.args.emb_dim)
        self.rel_emb = nn.Embedding(self.args.num_rel, self.args.emb_dim)  #关系的rel emb
        self.rel_reverse_emb = nn.Embedding(self.args.num_rel, self.args.emb_dim)  #reverse关系的rel emb 
        self.h_weighted_vector = nn.Embedding(self.args.num_rel, self.args.emb_dim)   #interaction mactrix
        self.t_weighted_vector = nn.Embedding(self.args.num_rel, self.args.emb_dim)   #interaction mactrix
        # self.bias = nn.Embedding(2, self.args.emb_dim)
        self.hr_bias = nn.Parameter(torch.zeros([self.args.emb_dim]))
        self.tr_bias = nn.Parameter(torch.zeros([self.args.emb_dim]))
        sqrt_size = 6.0 / math.sqrt(self.args.emb_dim)
        nn.init.uniform_(tensor=self.ent_emb.weight.data, a=-sqrt_size, b=sqrt_size)
        nn.init.uniform_(tensor=self.rel_emb.weight.data, a=-sqrt_size, b=sqrt_size)
        nn.init.uniform_(tensor=self.rel_reverse_emb.weight.data, a=-sqrt_size, b=sqrt_size)
        nn.init.uniform_(tensor=self.h_weighted_vector.weight.data, a=-sqrt_size, b=sqrt_size)
        nn.init.uniform_(tensor=self.t_weighted_vector.weight.data, a=-sqrt_size, b=sqrt_size)

    def score_func(self, ent_emb, rel_emb, weighted_vector, mode):
        """Calculating the score of triples.

        The formula for calculating the score is :math:`\sigma(tanh(c_r \circ h + c_r \circ h \circ r + b)t ^ T)`

        Args:
            ent_emb: entity embedding
            rel_emb: relation embedding
            weighted_vector: Interaction matrix for entities and relations
            mode: Choose head-predict or tail-predict.

        Returns:
            score: The score of triples.
        """
        x = ent_emb * weighted_vector + rel_emb * ent_emb * weighted_vector
        if mode == "tail_predict":
            x = torch.tanh(x + self.hr_bias)
        else:
            x = torch.tanh(x + self.tr_bias)
        x = self.dropout(x)
        x = torch.mm(x, self.ent_emb.weight.data.t())
        x = torch.sigmoid(x)
        return x

    def forward(self, triples, mode="single"):
        """The functions used in the training phase, calculate hr_score and tr_score simultaneously
        
        """
        head_emb = self.ent_emb(triples[:, 0])
        tail_emb = self.ent_emb(triples[:, 2])
        rel_emb = self.rel_emb(triples[:, 1])
        rel_reverse_emb = self.rel_reverse_emb(triples[:, 1])
        h_weighted_vector = self.h_weighted_vector(triples[:, 1])
        t_weighted_vector = self.t_weighted_vector(triples[:, 1])
        hr_score = self.score_func(head_emb, rel_emb, h_weighted_vector, "tail_predict")
        tr_score = self.score_func(tail_emb, rel_reverse_emb, t_weighted_vector, "head_predict")
        # bias = self.bias(triples_id)
        return hr_score, tr_score
    def get_score(self, batch, mode):
        """The functions used in the testing phase, predict triple score

        """
        triples = batch["positive_sample"]
        if mode == "tail_predict":
            head_emb = self.ent_emb(triples[:, 0])
            rel_emb = self.rel_emb(triples[:, 1])
            h_weighted_vector = self.h_weighted_vector(triples[:, 1])
            return self.score_func(head_emb, rel_emb, h_weighted_vector, "tail_predict")
        else:
            tail_emb = self.ent_emb(triples[:, 2])
            rel_reverse_emb = self.rel_reverse_emb(triples[:, 1])
            t_weighted_vector = self.t_weighted_vector(triples[:, 1])
            return self.score_func(tail_emb, rel_reverse_emb, t_weighted_vector, "head_predict")

    def regularize_loss(self, norm=2):
        """Add regularization to loss

        """
        return (self.ent_emb.weight.norm(p = norm) ** norm + \
            self.rel_emb.weight.norm(p = norm) ** norm + \
            self.rel_reverse_emb.weight.norm(p = norm) ** norm + \
            self.h_weighted_vector.weight.norm(p = norm) ** norm + \
            self.t_weighted_vector.weight.norm(p = norm) ** norm + \
            self.hr_bias.norm(p = norm) ** norm + \
            self.tr_bias.norm(p=norm) ** norm)

    
   




from .Adv_Loss import Adv_Loss
from .ComplEx_NNE_AER_Loss import ComplEx_NNE_AER_Loss
from .SimplE_Loss import SimplE_Loss
from .Cross_Entropy_Loss import Cross_Entropy_Loss
from .RGCN_Loss import RGCN_Loss
from .KBAT_Loss import KBAT_Loss
from .CrossE_Loss import CrossE_Loss
from .IterE_Loss import IterE_Loss
from .Margin_Loss import Margin_Loss
from .RugE_Loss import RugE_Loss


import torch
import torch.nn.functional as F
import torch.nn as nn

class KBAT_Loss(nn.Module):

    def __init__(self, args, model):
        super(KBAT_Loss, self).__init__()
        self.args = args
        self.model = model
        self.GAT_loss = nn.MarginRankingLoss(self.args.margin)
        self.Con_loss = nn.SoftMarginLoss()
      
    def forward(self, model, score, neg_score=None, label=None):
        if model == 'GAT':
            y     = -torch.ones( 2 * self.args.num_neg * self.args.train_bs).type_as(score)
            score = torch.tile(score, (2*self.args.num_neg, 1)).reshape(-1)
            loss  = self.GAT_loss(score, neg_score, y)
        elif model == 'ConvKB':
            loss = self.Con_loss(score.view(-1), label.view(-1))

        return loss


"""Base DataModule class."""
from pathlib import Path
from typing import Dict
import argparse
import os

import pytorch_lightning as pl
from torch.utils.data import DataLoader


class Config(dict):
    def __getattr__(self, name):
        return self.get(name)

    def __setattr__(self, name, val):
        self[name] = val


BATCH_SIZE = 8
NUM_WORKERS = 8


class BaseDataModule(pl.LightningDataModule):
    """
    Base DataModule.
    Learn more at https://pytorch-lightning.readthedocs.io/en/stable/datamodules.html
    """

    def __init__(self, args) -> None:
        super().__init__()
        self.args = args


    @staticmethod
    def add_to_argparse(parser):
        
        parser.add_argument(
            "--train_bs",
            type=int,
            default=0,
            help="Number of examples to operate on per forward step.",
        )
        parser.add_argument(
            "--num_batches",
            type=int,
            default=0,
            help="Number of examples to operate on per forward step.",
        )
        parser.add_argument(
            "--eval_bs",
            type=int,
            default=16,
            help="Number of examples to operate on per forward step.",
        )
        parser.add_argument(
            "--num_workers",
            type=int,
            default=8,
            help="Number of additional processes to load data.",
        )
        parser.add_argument(
            "--data_path",
            type=str,
            default="./dataset/WN18RR",
            help="Number of additional processes to load data.",
        )
        return parser

    def prepare_data(self):
        """
        Use this method to do things that might write to disk or that need to be done only from a single GPU in distributed settings (so don't set state `self.x = y`).
        """
        pass

    def setup(self, stage=None):
        """
        Split into train, val, test, and set dims.
        Should assign `torch Dataset` objects to self.data_train, self.data_val, and optionally self.data_test.
        """
        self.data_train = None
        self.data_val = None
        self.data_test = None

    def train_dataloader(self):
        return DataLoader(self.data_train, shuffle=True, batch_size=self.batch_size, num_workers=self.num_workers, pin_memory=True)

    def val_dataloader(self):
        return DataLoader(self.data_val, shuffle=False, batch_size=self.batch_size, num_workers=self.num_workers, pin_memory=True)

    def test_dataloader(self):
        return DataLoader(self.data_test, shuffle=False, batch_size=self.batch_size, num_workers=self.num_workers, pin_memory=True)

    def get_config(self):
        return dict(num_labels=self.num_labels)


import torch
import torch.nn as nn
from .model import Model
from IPython import embed
from torch.autograd import Variable


class ConvE(Model):

    """`Convolutional 2D Knowledge Graph Embeddings`_ (ConvE), which use a 2D convolution network for embedding representation.

    Attributes:
        args: Model configuration parameters.
    
    .. _Convolutional 2D Knowledge Graph Embeddings: https://arxiv.org/pdf/1707.01476.pdf
    """

    def __init__(self, args):
        super(ConvE, self).__init__(args)
        self.args = args
        self.emb_ent = None
        self.emb_rel = None

        self.init_emb(args)

    def init_emb(self,args):

        """Initialize the convolution layer and embeddings .

        Args:
            conv1: The convolution layer.
            fc: The full connection layer.
            bn0, bn1, bn2: The batch Normalization layer.
            inp_drop, hid_drop, feg_drop: The dropout layer.
            emb_ent: Entity embedding, shape:[num_ent, emb_dim].
            emb_rel: Relation_embedding, shape:[num_rel, emb_dim].
        """

        self.emb_dim1 = self.args.emb_shape
        self.emb_dim2 = self.args.emb_dim // self.emb_dim1
        self.emb_ent = torch.nn.Embedding(self.args.num_ent, self.args.emb_dim, padding_idx=0)
        self.emb_rel = torch.nn.Embedding(self.args.num_rel, self.args.emb_dim, padding_idx=0)
        torch.nn.init.xavier_normal_(self.emb_ent.weight.data)
        torch.nn.init.xavier_normal_(self.emb_rel.weight.data)
        # Setting dropout
        self.inp_drop = torch.nn.Dropout(self.args.inp_drop)
        self.hid_drop = torch.nn.Dropout(self.args.hid_drop)
        self.feg_drop = torch.nn.Dropout2d(self.args.fet_drop)
        # Setting net model
        self.conv1 = torch.nn.Conv2d(1, 32, (3, 3), 1, 0, bias=True)
        self.bn0 = torch.nn.BatchNorm2d(1)
        self.bn1 = torch.nn.BatchNorm2d(32)
        self.bn2 = torch.nn.BatchNorm1d(self.args.emb_dim)
        self.register_parameter('b', torch.nn.Parameter(torch.zeros(self.args.num_ent)))
        self.fc = torch.nn.Linear(self.args.hid_size,self.args.emb_dim)

    def score_func(self, head_emb, relation_emb, choose_emb = None):

        """Calculate the score of the triple embedding.

        This function calculate the score of the embedding.
        First, the entity and relation embeddings are reshaped
        and concatenated; the resulting matrix is then used as
        input to a convolutional layer; the resulting feature
        map tensor is vectorised and projected into a k-dimensional
        space.

        Args:
            head_emb: The embedding of head entity.
            relation_emb:The embedding of relation.

        Returns:
            score: Final score of the embedding.
        """

        stacked_inputs = torch.cat([head_emb, relation_emb], 2)
        stacked_inputs = self.bn0(stacked_inputs)
        x = self.inp_drop(stacked_inputs)
        x = self.conv1(x)
        x = self.bn1(x)
        x = torch.nn.functional.relu(x)
        x = self.feg_drop(x)
        x = x.view(x.shape[0], -1)
        x = self.fc(x)
        x = self.hid_drop(x)
        x = self.bn2(x)
        x = torch.nn.functional.relu(x)
        x = torch.mm(x, self.emb_ent.weight.transpose(1,0)) if choose_emb == None \
            else torch.mm(x, choose_emb.transpose(1, 0)) 
        x += self.b.expand_as(x)
        x = torch.sigmoid(x)
        return x

    def forward(self, triples):

        """The functions used in the training phase

        Args:
            triples: The triples ids, as (h, r, t), shape:[batch_size, 3].

        Returns:
            score: The score of triples.
        """
        
        head_emb = self.emb_ent(triples[:, 0]).view(-1, 1, self.emb_dim1, self.emb_dim2)
        rela_emb = self.emb_rel(triples[:, 1]).view(-1, 1, self.emb_dim1, self.emb_dim2)
        score = self.score_func(head_emb, rela_emb)
        return score

    def get_score(self, batch):

        """The functions used in the testing phase

        Args:
            batch: A batch of data.

        Returns:
            score: The score of triples.
        """
        
        triples = batch["positive_sample"]
        head_emb = self.emb_ent(triples[:, 0]).view(-1, 1, self.emb_dim1, self.emb_dim2)
        rela_emb = self.emb_rel(triples[:, 1]).view(-1, 1, self.emb_dim1, self.emb_dim2)
        score = self.score_func(head_emb, rela_emb)
        return score





import torch.nn as nn
import torch
import os
from .model import Model
from IPython import embed
from collections import defaultdict
import numpy as np
import pickle 
import copy

class IterE(Model):
    """`Iteratively Learning Embeddings and Rules for Knowledge Graph Reasoning. (WWW'19)`_ (IterE).

    Attributes:
        args: Model configuration parameters.
        epsilon: Caculate embedding_range.
        margin: Caculate embedding_range and loss.
        embedding_range: Uniform distribution range.
        ent_emb: Entity embedding, shape:[num_ent, emb_dim].
        rel_emb: Relation_embedding, shape:[num_rel, emb_dim].
    
    .. _Iteratively Learning Embeddings and Rules for Knowledge Graph Reasoning. (WWW'19): https://dl.acm.org/doi/10.1145/3308558.3313612
    """

    def __init__(self, args, train_sampler, test_sampler):
        super(IterE, self).__init__(args)
        self.args = args
        self.ent_emb = None
        self.rel_emb = None
        self.init_emb()
        #print(self.args)
        #print(train_sampler)
        #print('run get_axiom()')
        self.train_sampler = train_sampler
        self.train_triples_base = copy.deepcopy(train_sampler.train_triples)
        self.select_probability = self.args.select_probability
        self.max_entialments = self.args.max_entialments
        self.axiom_types  = self.args.axiom_types
        self.axiom_weight = self.args.axiom_weight
        self.inject_triple_percent = self.args.inject_triple_percent
        self.sparsity = 0.995
        
        self.num_entity = self.args.num_ent
        
        self.relation2id=train_sampler.rel2id
        
        self.train_ids=train_sampler.train_triples
        self.valid_ids=train_sampler.valid_triples
        self.test_ids=train_sampler.test_triples
        
        #print(len(self.train_ids))
        #print(len(self.valid_ids))
        #print(len(self.test_ids))
        
        self.train_ids_labels_inject = np.reshape([], [-1, 4])
        
        # generate r_ht, hr_t
        print('# generate r_ht, hr_t')
        self.r_ht, self.hr_t, self.tr_h, self.hr_t_all, self.tr_h_all = self._generate(self.train_ids, self.valid_ids, self.test_ids)
        
        # generate entity2frequency and entity2sparsity dict
        print('# generate entity2frequency and entity2sparsity dict')
        self.entity2frequency, self.entity2sparsity = self._entity2frequency()
    
        print('# get_axiom')
        self.get_axiom()
        
        
        #self.rule, self.conf = self.get_rule(self.relation2id)
        
    def _entity2frequency(self):
        ent2freq = {ent:0 for ent in range(self.num_entity)}
        ent2sparsity = {ent:-1 for ent in range(self.num_entity)}
        for h,r,t in self.train_ids:
            ent2freq[h] += 1
            ent2freq[t] += 1
        ent_freq_list = np.asarray([ent2freq[ent] for ent in range(self.num_entity)])
        ent_freq_list_sort = np.argsort(ent_freq_list)

        max_freq = max(list(ent2freq))
        min_freq = min(list(ent2freq))
        for ent, freq in ent2freq.items():
            sparsity = 1 - (freq-min_freq)/(max_freq - min_freq)
            ent2sparsity[ent] = sparsity
        return ent2freq, ent2sparsity

    
    def _generate(self, train, valid, test):
        r_ht = defaultdict(set)
        hr_t = defaultdict(set)
        tr_h = defaultdict(set)
        hr_t_all = defaultdict(list)
        tr_h_all = defaultdict(list)
        for (h,r,t) in train:
            r_ht[r].add((h,t))
            hr_t[(h,r)].add(t)
            tr_h[(t,r)].add(h)
            hr_t_all[(h,r)].append(t)
            tr_h_all[(t,r)].append(h)
        for (h,r,t) in test+valid:
            hr_t_all[(h,r)].append(t)
            tr_h_all[(t, r)].append(h)
        return r_ht, hr_t, tr_h, hr_t_all, tr_h_all

    def get_axiom(self, ):
        self.axiom_dir = os.path.join(self.args.data_path, 'axiom_pool')
        self.reflexive_dir, self.symmetric_dir, self.transitive_dir, self.inverse_dir, self.subproperty_dir, self.equivalent_dir, self.inferencechain1, self.inferencechain2, self.inferencechain3, self.inferencechain4 = map(lambda x: os.path.join(self.axiom_dir, x),
                                                             ['axiom_reflexive.txt',
                                                              'axiom_symmetric.txt',
                                                              'axiom_transitive.txt',
                                                              'axiom_inverse.txt',
                                                              'axiom_subProperty.txt',
                                                              'axiom_equivalent.txt',
                                                              'axiom_inferenceChain1.txt',
                                                              'axiom_inferenceChain2.txt',
                                                              'axiom_inferenceChain3.txt',
                                                              'axiom_inferenceChain4.txt'])
        # read and materialize axioms
        print('# self._read_axioms()')
        self._read_axioms()
        print('# self._read_axioms()')
        self._materialize_axioms()
        print('# self._read_axioms()')
        self._init_valid_axioms()
        
        
        
        
    def _read_axioms(self):
        # for each axiom, the first id is the basic relation
        self.axiompool_reflexive = self._read_axiompool_file(self.reflexive_dir)
        self.axiompool_symmetric = self._read_axiompool_file(self.symmetric_dir)
        self.axiompool_transitive = self._read_axiompool_file(self.transitive_dir)
        self.axiompool_inverse = self._read_axiompool_file(self.inverse_dir)
        self.axiompool_equivalent = self._read_axiompool_file(self.equivalent_dir)
        self.axiompool_subproperty = self._read_axiompool_file(self.subproperty_dir)
        self.axiompool_inferencechain1 = self._read_axiompool_file(self.inferencechain1)
        self.axiompool_inferencechain2 = self._read_axiompool_file(self.inferencechain2)
        self.axiompool_inferencechain3 = self._read_axiompool_file(self.inferencechain3)
        self.axiompool_inferencechain4 = self._read_axiompool_file(self.inferencechain4)
        self.axiompool = [self.axiompool_reflexive, self.axiompool_symmetric, self.axiompool_transitive,
                          self.axiompool_inverse, self.axiompool_subproperty, self.axiompool_equivalent,
                          self.axiompool_inferencechain1,self.axiompool_inferencechain2,
                          self.axiompool_inferencechain3,self.axiompool_inferencechain4]
    
    def _read_axiompool_file(self, file):
        f = open(file, 'r')
        axioms = []
        for line in f.readlines():
            line_list = line.strip().split('\t')
            axiom_ids = list(map(lambda x: self.relation2id[x], line_list))
            #axiom_ids = self.relation2id[line_list]
            axioms.append(axiom_ids)
        # for the case reflexive pool is empty
        if len(axioms) == 0:
            np.reshape(axioms, [-1, 3])
        return axioms
    
    # for each axioms in axiom pool
    # generate a series of entailments for each axiom
    def _materialize_axioms(self, generate=True, dump=True, load=False):
        if generate:
            self.reflexive2entailment = defaultdict(list)
            self.symmetric2entailment = defaultdict(list)
            self.transitive2entailment = defaultdict(list)
            self.inverse2entailment = defaultdict(list)
            self.equivalent2entailment = defaultdict(list)
            self.subproperty2entailment = defaultdict(list)
            self.inferencechain12entailment = defaultdict(list)
            self.inferencechain22entailment = defaultdict(list)
            self.inferencechain32entailment = defaultdict(list)
            self.inferencechain42entailment = defaultdict(list)

            self.reflexive_entailments, self.reflexive_entailments_num = self._materialize_sparse(self.axiompool_reflexive, type='reflexive')
            self.symmetric_entailments, self.symmetric_entailments_num = self._materialize_sparse(self.axiompool_symmetric, type='symmetric')
            self.transitive_entailments, self.transitive_entailments_num = self._materialize_sparse(self.axiompool_transitive, type='transitive')
            self.inverse_entailments, self.inverse_entailments_num = self._materialize_sparse(self.axiompool_inverse, type='inverse')
            self.subproperty_entailments, self.subproperty_entailments_num = self._materialize_sparse(self.axiompool_subproperty, type='subproperty')
            self.equivalent_entailments, self.equivalent_entailments_num  = self._materialize_sparse(self.axiompool_equivalent, type='equivalent')

            self.inferencechain1_entailments, self.inferencechain1_entailments_num = self._materialize_sparse(self.axiompool_inferencechain1, type='inferencechain1')
            self.inferencechain2_entailments, self.inferencechain2_entailments_num = self._materialize_sparse(self.axiompool_inferencechain2, type='inferencechain2')
            self.inferencechain3_entailments, self.inferencechain3_entailments_num = self._materialize_sparse(self.axiompool_inferencechain3, type='inferencechain3')
            self.inferencechain4_entailments, self.inferencechain4_entailments_num = self._materialize_sparse(self.axiompool_inferencechain4, type='inferencechain4')


            print('reflexive entailments for sparse: ', self.reflexive_entailments_num)
            print('symmetric entailments for sparse: ', self.symmetric_entailments_num)
            print('transitive entailments for sparse: ', self.transitive_entailments_num)
            print('inverse entailments for sparse: ', self.inverse_entailments_num)
            print('subproperty entailments for sparse: ', self.subproperty_entailments_num)
            print('equivalent entailments for sparse: ', self.equivalent_entailments_num)
            print('inferencechain1 entailments for sparse: ', self.inferencechain1_entailments_num)
            print('inferencechain2 entailments for sparse: ', self.inferencechain2_entailments_num)
            print('inferencechain3 entailments for sparse: ', self.inferencechain3_entailments_num)
            print('inferencechain4 entailments for sparse: ', self.inferencechain4_entailments_num)


            print("finish generate axioms entailments for sparse")


        if dump:
            pickle.dump(self.reflexive_entailments, open(os.path.join(self.axiom_dir, 'reflexive_entailments'), 'wb'))
            pickle.dump(self.symmetric_entailments, open(os.path.join(self.axiom_dir, 'symmetric_entailments'), 'wb'))
            pickle.dump(self.transitive_entailments, open(os.path.join(self.axiom_dir, 'transitive_entailments'), 'wb'))
            pickle.dump(self.inverse_entailments, open(os.path.join(self.axiom_dir, 'inverse_entailments'), 'wb'))
            pickle.dump(self.subproperty_entailments, open(os.path.join(self.axiom_dir, 'subproperty_entailments'), 'wb'))
            #pickle.dump(self.inferencechain_entailments, open(os.path.join(self.axiom_dir, 'inferencechain_entailments'), 'wb'))
            pickle.dump(self.equivalent_entailments, open(os.path.join(self.axiom_dir, 'equivalent_entailments'), 'wb'))

            pickle.dump(self.inferencechain1_entailments,
                        open(os.path.join(self.axiom_dir, 'inferencechain1_entailments'), 'wb'))
            pickle.dump(self.inferencechain2_entailments,
                        open(os.path.join(self.axiom_dir, 'inferencechain2_entailments'), 'wb'))
            pickle.dump(self.inferencechain3_entailments,
                        open(os.path.join(self.axiom_dir, 'inferencechain3_entailments'), 'wb'))
            pickle.dump(self.inferencechain4_entailments,
                        open(os.path.join(self.axiom_dir, 'inferencechain4_entailments'), 'wb'))

            print("finish dump axioms entialments")

        if load:
            print("load refexive entailments...")
            self.reflexive_entailments = pickle.load(open(os.path.join(self.axiom_dir, 'reflexive_entailments'), 'rb'))
            print(self.reflexive_entailments)
            print('load symmetric entailments...')
            self.symmetric_entailments = pickle.load(open(os.path.join(self.axiom_dir, 'symmetric_entailments'), 'rb'))
            print("load transitive entialments... ")
            self.transitive_entailments = pickle.load(open(os.path.join(self.axiom_dir, 'transitive_entailments'), 'rb'))
            print("load inverse entailments...")
            self.inverse_entailments = pickle.load(open(os.path.join(self.axiom_dir, 'inverse_entailments'), 'rb'))
            print("load subproperty entailments...")
            self.subproperty_entailments = pickle.load(open(os.path.join(self.axiom_dir, 'subproperty_entailments'), 'rb'))
            #print("load inferencechain entailments...")
            #self.inferencechain_entailments = pickle.load(open(os.path.join(self.axiom_dir, 'inferencechain_entailments'), 'rb'))
            print("load equivalent entialments...")
            self.equivalent_entailments = pickle.load(open(os.path.join(self.axiom_dir, 'equivalent_entailments'), 'rb'))

            print("load inferencechain1 entailments...")
            self.inferencechain1_entailments = pickle.load(
                open(os.path.join(self.axiom_dir, 'inferencechain1_entailments'), 'rb'))
            print("load inferencechain2 entailments...")
            self.inferencechain2_entailments = pickle.load(
                open(os.path.join(self.axiom_dir, 'inferencechain2_entailments'), 'rb'))
            print("load inferencechain3 entailments...")
            self.inferencechain3_entailments = pickle.load(
                open(os.path.join(self.axiom_dir, 'inferencechain3_entailments'), 'rb'))
            print("load inferencechain4 entailments...")
            self.inferencechain4_entailments = pickle.load(
                open(os.path.join(self.axiom_dir, 'inferencechain4_entailments'), 'rb'))

            print("finish load axioms entailments")

    def _materialize_sparse(self, axioms, type=None, sparse = False):
        inference = []
        # axiom2entailment is a dict
        # with the all axioms in the axiom pool as keys
        # and all the entailments for each axiom as values
        axiom_list = axioms
        length = len(axioms)
        max_entailments = self.max_entialments
        num = 0
        if length == 0:
            if type == 'reflexive':
                np.reshape(inference, [-1, 3])
            elif type == 'symmetric' or type =='inverse' or  type =='equivalent' or type =='subproperty':
                np.reshape(inference, [-1, 6])
            elif type=='transitive' or type=='inferencechain':
                np.reshape(inference, [-1, 9])
            else:
                raise NotImplementedError
            return inference, num

        if type == 'reflexive':
            for axiom in axiom_list:
                axiom_key =tuple(axiom)
                r = axiom[0]
                inference_tmp = []
                for (h,t) in self.r_ht[r]:
                    # filter the axiom with too much entailments
                    if len(inference_tmp) > max_entailments:
                        inference_tmp = np.reshape([], [-1, 2])
                        break
                    if h != t and self.entity2sparsity[h]>self.sparsity:
                        num += 1
                        inference_tmp.append([h,r,h])

                for entailment in inference_tmp:
                    self.reflexive2entailment[axiom_key].append(entailment)
                inference.append(inference_tmp)

        if type == 'symmetric':
            #self.symmetric2entailment = defaultdict(list)
            for axiom in axiom_list:
                axiom_key = tuple(axiom)
                r = axiom[0]
                inference_tmp = []
                for (h,t) in self.r_ht[r]:
                    if len(inference_tmp) > max_entailments:
                        inference_tmp = np.reshape([], [-1, 2])
                        break
                    if (t,h) not in self.r_ht[r] and (self.entity2sparsity[h]>self.sparsity or self.entity2sparsity[t]>self.sparsity):
                        num += 1
                        inference_tmp.append([h,r,t,t,r,h])


                for entailment in inference_tmp:
                    self.symmetric2entailment[axiom_key].append(entailment)
                inference.append(inference_tmp)

        if type == 'transitive':
            #self.transitive2entailment = defaultdict(list)
            for axiom in axiom_list:
                axiom_key = tuple(axiom)
                r = axiom[0]
                inference_tmp = []
                for (h,t) in self.r_ht[r]:
                    if len(inference_tmp) > max_entailments:
                        inference_tmp = np.reshape([], [-1, 9])
                        break
                    # (t,r,e) exist but (h,r,e) not exist and e!=h
                    for e in self.hr_t[(t,r)]- self.hr_t[(h,r)]:
                        if e != h and (self.entity2sparsity[h]>self.sparsity or self.entity2sparsity[e]>self.sparsity):
                            num += 1
                            inference_tmp.append([h,r,t,t,r,e,h,r,e])

                for entailment in inference_tmp:
                    self.transitive2entailment[axiom_key].append(entailment)
                inference.append(inference_tmp)

        if type == 'inverse':
            for axiom in axiom_list:
                axiom_key = tuple(axiom)
                r1,r2 = axiom
                inference_tmp = []
                for (h,t) in self.r_ht[r1]:
                    if len(inference_tmp) > max_entailments:
                        inference_tmp = np.reshape([], [-1, 6])
                        break
                    if (t,h) not in self.r_ht[r2] and (self.entity2sparsity[h]>self.sparsity or self.entity2sparsity[t]>self.sparsity):
                        num += 1
                        inference_tmp.append([h,r1,t, t,r2,h])
                        #self.inverse2entailment[axiom_key].append([h,r1,t, t,r2,h])

                for entailment in inference_tmp:
                    self.inverse2entailment[axiom_key].append(entailment)
                inference.append(inference_tmp)

        if type == 'equivalent' or type =='subproperty':
            for axiom in axiom_list:
                axiom_key = tuple(axiom)
                r1,r2 = axiom
                inference_tmp = []
                for (h,t) in self.r_ht[r1]:
                    if len(inference_tmp) > max_entailments:
                        inference_tmp = np.reshape([], [-1, 6])
                        break
                    if (h,t) not in self.r_ht[r2] and (self.entity2sparsity[h]>self.sparsity or self.entity2sparsity[t]>self.sparsity):
                        num += 1
                        inference_tmp.append([h,r1,t, h,r2,t])

                for entailment in inference_tmp:
                    self.equivalent2entailment[axiom_key].append(entailment)
                    self.subproperty2entailment[axiom_key].append(entailment)
                inference.append(inference_tmp)


        if type == 'inferencechain1':
            self.inferencechain12entailment = defaultdict(list)
            i = 0
            for axiom in axiom_list:
                axiom_key = tuple(axiom)
                i += 1
                # print('%d/%d' % (i, length))
                r1, r2, r3 = axiom
                inference_tmp = []
                for (e, h) in self.r_ht[r2]:
                    if len(inference_tmp) > max_entailments:
                        inference_tmp = np.reshape([], [-1, 9])
                        break
                    for t in self.hr_t[(e, r3)]:
                        if (h, t) not in self.r_ht[r1] and (
                                        self.entity2sparsity[h] > self.sparsity or self.entity2sparsity[e] > self.sparsity):
                            num += 1
                            inference_tmp.append([e, r2, h, e, r3, t, h, r1, t])
                            #self.inferencechain12entailment[axiom_key].append([[e, r2, h, e, r3, t, h, r1, t]])


                for entailment in inference_tmp:
                    self.inferencechain12entailment[axiom_key].append(entailment)
                inference.append(inference_tmp)

        if type == 'inferencechain2':
            self.inferencechain22entailment = defaultdict(list)
            i = 0
            for axiom in axiom_list:
                axiom_key = tuple(axiom)
                i += 1
                # print('%d/%d' % (i, length))
                r1, r2, r3 = axiom
                inference_tmp = []
                for (e, h) in self.r_ht[r2]:
                    if len(inference_tmp) > max_entailments:
                        inference_tmp = np.reshape([], [-1, 9])
                        break
                    for t in self.tr_h[(e, r3)]:
                        if (h, t) not in self.r_ht[r1] and (
                                        self.entity2sparsity[h] > self.sparsity or self.entity2sparsity[e] > self.sparsity):
                            num += 1
                            inference_tmp.append([e, r2, h, t, r3, e, h, r1, t])
                            #self.inferencechain22entailment[axiom_key].append([[e, r2, h, t, r3, e, h, r1, t]])

                for entailment in inference_tmp:
                    self.inferencechain22entailment[axiom_key].append(entailment)
                inference.append(inference_tmp)


        if type == 'inferencechain3':
            self.inferencechain32entailment = defaultdict(list)
            i = 0
            for axiom in axiom_list:
                axiom_key = tuple(axiom)
                i += 1
                # print('%d/%d' % (i, length))
                r1, r2, r3 = axiom
                inference_tmp = []
                for (h, e) in self.r_ht[r2]:
                    if len(inference_tmp) > max_entailments:
                        inference_tmp = np.reshape([], [-1, 9])
                        break
                    for t in self.hr_t[(e, r3)]:
                        if (h, t) not in self.r_ht[r1] and (
                                        self.entity2sparsity[h] > self.sparsity or self.entity2sparsity[e] > self.sparsity):
                            num += 1
                            inference_tmp.append([h, r2, e, e, r3, t, h, r1, t])


                for entailment in inference_tmp:
                    self.inferencechain32entailment[axiom_key].append(entailment)
                inference.append(inference_tmp)

        if type == 'inferencechain4':
            self.inferencechain42entailment = defaultdict(list)
            i = 0
            for axiom in axiom_list:
                axiom_key = tuple(axiom)
                i += 1
                # print('%d/%d' % (i, length))
                r1, r2, r3 = axiom
                inference_tmp = []
                for (h, e) in self.r_ht[r2]:
                    if len(inference_tmp) > max_entailments:
                        inference_tmp = np.reshape([], [-1, 9])
                        break
                    for t in self.tr_h[(e, r3)]:
                        if (h, t) not in self.r_ht[r1] and (
                                        self.entity2sparsity[h] > self.sparsity or self.entity2sparsity[e] > self.sparsity):
                            num += 1
                            inference_tmp.append([h, r2, e, t, r3, e, h, r1, t])

                for entailment in inference_tmp:
                    self.inferencechain42entailment[axiom_key].append(entailment)
                inference.append(inference_tmp)
        return inference, num

    def _materialize(self, axioms, type=None, sparse=False):
        inference = []
        # axiom2entailment is a dict
        # with the all axioms in the axiom pool as keys
        # and all the entailments for each axiom as values
        axiom_list = axioms
        # print('axiom_list', axiom_list)
        length = len(axioms)
        max_entailments = 5000
        num = 0
        if length == 0:
            if type == 'reflexive':
                np.reshape(inference, [-1, 3])
            elif type == 'symmetric' or type == 'inverse' or type == 'equivalent' or type == 'subproperty':
                np.reshape(inference, [-1, 6])
            elif type == 'transitive' or type == 'inferencechain':
                np.reshape(inference, [-1, 9])
            else:
                raise NotImplementedError
            return inference, num

        if type == 'reflexive':
            for axiom in axiom_list:
                axiom_key = tuple(axiom)
                r = axiom[0]
                inference_tmp = []
                for (h, t) in self.r_ht[r]:
                    if len(inference_tmp) > max_entailments:
                        inference_tmp = np.reshape([], [-1, 2])
                        break
                    if h != t: #and self.entity2sparsity[h] > self.sparsity:
                        num += 1
                        inference_tmp.append([h, r, h])
                for entailment in inference_tmp:
                    self.reflexive2entailment[axiom_key].append(entailment)
                inference.append(inference_tmp)

        if type == 'symmetric':
            for axiom in axiom_list:
                axiom_key = tuple(axiom)
                r = axiom[0]
                inference_tmp = []
                for (h, t) in self.r_ht[r]:
                    if len(inference_tmp) > max_entailments:
                        inference_tmp = np.reshape([], [-1, 2])
                        break
                    if (t, h) not in self.r_ht[r]: #and (self.entity2sparsity[h] > self.sparsity or self.entity2sparsity[t] > self.sparsity):
                        num += 1
                        inference_tmp.append([h, r, t, t, r, h])
                for entailment in inference_tmp:
                    self.symmetric2entailment[axiom_key].append(entailment)
                inference.append(inference_tmp)

        if type == 'transitive':
            for axiom in axiom_list:
                axiom_key = tuple(axiom)
                r = axiom[0]
                inference_tmp = []
                for (h, t) in self.r_ht[r]:
                    if len(inference_tmp) > max_entailments:
                        inference_tmp = np.reshape([], [-1, 9])
                        break
                    # (t,r,e) exist but (h,r,e) not exist and e!=h
                    for e in self.hr_t[(t, r)] - self.hr_t[(h, r)]:
                        if e != h: #and (self.entity2sparsity[h] > self.sparsity or self.entity2sparsity[e] > self.sparsity):
                            num += 1
                            inference_tmp.append([h, r, t, t, r, e, h, r, e])

                for entailment in inference_tmp:
                    self.transitive2entailment[axiom_key].append(entailment)
                inference.append(inference_tmp)

        if type == 'inverse':
            # self.inverse2entailment = defaultdict(list)
            for axiom in axiom_list:
                axiom_key = tuple(axiom)
                r1, r2 = axiom
                inference_tmp = []
                for (h, t) in self.r_ht[r1]:
                    if len(inference_tmp) > max_entailments:
                        inference_tmp = np.reshape([], [-1, 6])
                        break
                    if (t, h) not in self.r_ht[r2]: #and (self.entity2sparsity[h] > self.sparsity or self.entity2sparsity[t] > self.sparsity):
                        num += 1
                        inference_tmp.append([h, r1, t, t, r2, h])
                for entailment in inference_tmp:
                    self.inverse2entailment[axiom_key].append(entailment)
                inference.append(inference_tmp)

        if type == 'equivalent' or type == 'subproperty':
            for axiom in axiom_list:
                axiom_key = tuple(axiom)
                r1, r2 = axiom
                inference_tmp = []
                for (h, t) in self.r_ht[r1]:
                    if len(inference_tmp) > max_entailments:
                        inference_tmp = np.reshape([], [-1, 6])
                        break
                    if (h, t) not in self.r_ht[r2]: #and (self.entity2sparsity[h] > self.sparsity or self.entity2sparsity[t] > self.sparsity):
                        num += 1
                        inference_tmp.append([h, r1, t, h, r2, t])

                for entailment in inference_tmp:
                    self.equivalent2entailment[axiom_key].append(entailment)
                    self.subproperty2entailment[axiom_key].append(entailment)
                inference.append(inference_tmp)

        if type == 'inferencechain1':
            self.inferencechain12entailment = defaultdict(list)
            i = 0
            for axiom in axiom_list:
                axiom_key = tuple(axiom)
                i += 1
                # print('%d/%d' % (i, length))
                r1, r2, r3 = axiom
                inference_tmp = []
                for (e, h) in self.r_ht[r2]:
                    if len(inference_tmp) > max_entailments:
                        inference_tmp = np.reshape([], [-1, 9])
                        break
                    for t in self.hr_t[(e, r3)]:
                        if (h, t) not in self.r_ht[r1]:
                            num += 1
                            inference_tmp.append([e, r2, h, e, r3, t, h, r1, t])
                for entailment in inference_tmp:
                    self.inferencechain12entailment[axiom_key].append(entailment)
                inference.append(inference_tmp)

        if type == 'inferencechain2':
            self.inferencechain22entailment = defaultdict(list)
            i = 0
            for axiom in axiom_list:
                axiom_key = tuple(axiom)
                i += 1
                # print('%d/%d' % (i, length))
                r1, r2, r3 = axiom
                inference_tmp = []
                for (e, h) in self.r_ht[r2]:
                    if len(inference_tmp) > max_entailments:
                        inference_tmp = np.reshape([], [-1, 9])
                        break
                    for t in self.tr_h[(e, r3)]:
                        if (h, t) not in self.r_ht[r1]:
                            num += 1
                            inference_tmp.append([e, r2, h, t, r3, e, h, r1, t])
                for entailment in inference_tmp:
                    self.inferencechain22entailment[axiom_key].append(entailment)
                inference.append(inference_tmp)

        if type == 'inferencechain3':
            self.inferencechain32entailment = defaultdict(list)
            i = 0
            for axiom in axiom_list:
                axiom_key = tuple(axiom)
                i += 1
                # print('%d/%d' % (i, length))
                r1, r2, r3 = axiom
                inference_tmp = []
                for (h, e) in self.r_ht[r2]:
                    if len(inference_tmp) > max_entailments:
                        inference_tmp = np.reshape([], [-1, 9])
                        break
                    for t in self.hr_t[(e, r3)]:
                        if (h, t) not in self.r_ht[r1]:
                            num += 1
                            inference_tmp.append([h, r2, e, e, r3, t, h, r1, t])
                for entailment in inference_tmp:
                    self.inferencechain32entailment[axiom_key].append(entailment)
                inference.append(inference_tmp)

        if type == 'inferencechain4':
            self.inferencechain42entailment = defaultdict(list)
            i = 0
            for axiom in axiom_list:
                axiom_key = tuple(axiom)
                i += 1
                r1, r2, r3 = axiom
                inference_tmp = []
                for (h, e) in self.r_ht[r2]:
                    if len(inference_tmp) > max_entailments:
                        inference_tmp = np.reshape([], [-1, 9])
                        break
                    for t in self.tr_h[(e, r3)]:
                        if (h, t) not in self.r_ht[r1]:
                            num += 1
                            inference_tmp.append([h, r2, e, t, r3, e, h, r1, t])
                for entailment in inference_tmp:
                    self.inferencechain42entailment[axiom_key].append(entailment)
                inference.append(inference_tmp)
        return inference, num
    
    def _init_valid_axioms(self):
        # init valid axioms
        self.valid_reflexive, self.valid_symmetric, self.valid_transitive,\
        self.valid_inverse, self.valid_subproperty, self.valid_equivalent,\
        self.valid_inferencechain1, self.valid_inferencechain2, \
        self.valid_inferencechain3, self.valid_inferencechain4 = [[] for x in range(self.axiom_types)]

        # init valid axiom entailments
        self.valid_reflexive2entailment, self.valid_symmetric2entailment, self.valid_transitive2entailment, \
        self.valid_inverse2entailment, self.valid_subproperty2entailment, self.valid_equivalent2entailment, \
        self.valid_inferencechain12entailment, self.valid_inferencechain22entailment, \
        self.valid_inferencechain32entailment, self.valid_inferencechain42entailment = [[] for x in range(self.axiom_types)]

        # init valid axiom entailments probability
        self.valid_reflexive_p, self.valid_symmetric_p, self.valid_transitive_p, \
        self.valid_inverse_p, self.valid_subproperty_p, self.valid_equivalent_p, \
        self.valid_inferencechain1_p, self.valid_inferencechain2_p,\
        self.valid_inferencechain3_p, self.valid_inferencechain4_p= [[] for x in range(self.axiom_types)]

        # init valid axiom batchsize
        self.reflexive_batchsize = 1
        self.symmetric_batchsize = 1
        self.transitive_batchsize = 1
        self.inverse_batchsize = 1
        self.subproperty_batchsize = 1
        self.equivalent_batchsize = 1
        #self.inferencechain_batchsize = 1
        self.inferencechain1_batchsize = 1
        self.inferencechain2_batchsize = 1
        self.inferencechain3_batchsize = 1
        self.inferencechain4_batchsize = 1


    # add the new triples from axioms to training triple
    def update_train_triples(self, epoch=0, update_per = 10):
        """add the new triples from axioms to training triple

        Args:
            epoch (int, optional): epoch in training process. Defaults to 0.
            update_per (int, optional): Defaults to 10.

        Returns:
            updated_train_data: training triple after adding the new triples from axioms
        """
        reflexive_triples, symmetric_triples, transitive_triples, inverse_triples,\
            equivalent_triples, subproperty_triples, inferencechain1_triples, \
            inferencechain2_triples, inferencechain3_triples, inferencechain4_triples = [ np.reshape(np.asarray([]), [-1, 3]) for i in range(self.axiom_types)]
        reflexive_p, symmetric_p, transitive_p, inverse_p, \
            equivalent_p, subproperty_p, inferencechain1_p, \
            inferencechain2_p, inferencechain3_p, inferencechain4_p = [np.reshape(np.asarray([]), [-1, 1]) for i in
                                                                           range(self.axiom_types)]
        updated_train_data=None
        if epoch >= 20:
        #if True:
            print("len(self.valid_reflexive2entailment):", len(self.valid_reflexive2entailment))
            print("len(self.valid_symmetric2entailment):", len(self.valid_symmetric2entailment))
            print("len(self.valid_transitive2entailment)", len(self.valid_transitive2entailment))
            print("len(self.valid_inverse2entailment)", len(self.valid_inverse2entailment))
            print("len(self.valid_equivalent2entailment)", len(self.valid_equivalent2entailment))
            print("len(self.valid_subproperty2entailment)", len(self.valid_subproperty2entailment))

            valid_reflexive2entailment, valid_symmetric2entailment, valid_transitive2entailment,\
            valid_inverse2entailment, valid_equivalent2entailment, valid_subproperty2entailment, \
            valid_inferencechain12entailment, valid_inferencechain22entailment,\
            valid_inferencechain32entailment, valid_inferencechain42entailment = [[] for i in range(10)]

            if len(self.valid_reflexive2entailment)>0:
                valid_reflexive2entailment = np.reshape(np.asarray(self.valid_reflexive2entailment), [-1, 3])
                reflexive_triples = np.asarray(valid_reflexive2entailment)[:, -3:]
                reflexive_p = np.reshape(np.asarray(self.valid_reflexive_p),[-1,1])

            if len(self.valid_symmetric2entailment) > 0:
                valid_symmetric2entailment = np.reshape(np.asarray(self.valid_symmetric2entailment), [-1, 6])
                symmetric_triples = np.asarray(valid_symmetric2entailment)[:, -3:]
                symmetric_p = np.reshape(np.asarray(self.valid_symmetric_p),[-1,1])

            if len(self.valid_transitive2entailment) > 0:
                valid_transitive2entailment = np.reshape(np.asarray(self.valid_transitive2entailment), [-1, 9])
                transitive_triples = np.asarray(valid_transitive2entailment)[:, -3:]
                transitive_p = np.reshape(np.asarray(self.valid_transitive_p), [-1, 1])

            if len(self.valid_inverse2entailment) > 0:
                valid_inverse2entailment = np.reshape(np.asarray(self.valid_inverse2entailment), [-1, 6])
                inverse_triples = np.asarray(valid_inverse2entailment)[:, -3:]
                inverse_p = np.reshape(np.asarray(self.valid_inverse_p), [-1, 1])

            if len(self.valid_equivalent2entailment) > 0:
                valid_equivalent2entailment = np.reshape(np.asarray(self.valid_equivalent2entailment), [-1, 6])
                equivalent_triples = np.asarray(valid_equivalent2entailment)[:, -3:]
                equivalent_p = np.reshape(np.asarray(self.valid_equivalent_p), [-1, 1])

            if len(self.valid_subproperty2entailment) > 0:
                valid_subproperty2entailment = np.reshape(np.asarray(self.valid_subproperty2entailment), [-1, 6])
                subproperty_triples = np.asarray(valid_subproperty2entailment)[:, -3:]
                subproperty_p = np.reshape(np.asarray(self.valid_subproperty_p),[-1,1])

            if len(self.valid_inferencechain12entailment) > 0:
                valid_inferencechain12entailment = np.reshape(np.asarray(self.valid_inferencechain12entailment), [-1, 9])
                inferencechain1_triples = np.asarray(valid_inferencechain12entailment)[:, -3:]
                inferencechain1_p = np.reshape(np.asarray(self.valid_inferencechain1_p), [-1, 1])

            if len(self.valid_inferencechain22entailment) > 0:
                valid_inferencechain22entailment = np.reshape(np.asarray(self.valid_inferencechain22entailment), [-1, 9])
                inferencechain2_triples = np.asarray(valid_inferencechain22entailment)[:, -3:]
                inferencechain2_p = np.reshape(np.asarray(self.valid_inferencechain2_p), [-1, 1])

            if len(self.valid_inferencechain32entailment) > 0:
                valid_inferencechain32entailment = np.reshape(np.asarray(self.valid_inferencechain32entailment), [-1, 9])
                inferencechain3_triples = np.asarray(valid_inferencechain32entailment)[:, -3:]
                inferencechain3_p = np.reshape(np.asarray(self.valid_inferencechain3_p), [-1, 1])

            if len(self.valid_inferencechain42entailment) > 0:
                valid_inferencechain42entailment = np.reshape(np.asarray(self.valid_inferencechain42entailment), [-1, 9])
                inferencechain4_triples = np.asarray(valid_inferencechain42entailment)[:, -3:]
                inferencechain4_p = np.reshape(np.asarray(self.valid_inferencechain4_p), [-1, 1])

            # pickle.dump(self.reflexive_entailments, open(os.path.join(self.axiom_dir, 'reflexive_entailments'), 'wb'))
            # store all the injected triples
            entailment_all = (valid_reflexive2entailment, valid_symmetric2entailment, valid_transitive2entailment,
                     valid_inverse2entailment, valid_equivalent2entailment, valid_subproperty2entailment,
                     valid_inferencechain12entailment,valid_inferencechain22entailment,
                              valid_inferencechain32entailment,valid_inferencechain42entailment)
            pickle.dump(entailment_all, open(os.path.join(self.axiom_dir, 'valid_entailments.pickle'), 'wb'))



        train_inject_triples = np.concatenate([reflexive_triples, symmetric_triples, transitive_triples, inverse_triples,
                                                equivalent_triples, subproperty_triples, inferencechain1_triples,
                                               inferencechain2_triples,inferencechain3_triples,inferencechain4_triples],
                                                axis=0)

        train_inject_triples_p = np.concatenate([reflexive_p,symmetric_p, transitive_p, inverse_p,
                                               equivalent_p, subproperty_p, inferencechain1_p,
                                                 inferencechain2_p,inferencechain3_p,inferencechain4_p],
                                              axis=0)

        self.train_inject_triples = train_inject_triples
        inject_labels = np.reshape(np.ones(len(train_inject_triples)), [-1, 1]) * self.axiom_weight * train_inject_triples_p
        train_inject_ids_labels = np.concatenate([train_inject_triples, inject_labels],
                                                axis=1)


        self.train_ids_labels_inject = train_inject_triples#train_inject_ids_labels


        print('num reflexive triples', len(reflexive_triples))
        print('num symmetric triples', len(symmetric_triples))
        print('num transitive triples', len(transitive_triples))
        print('num inverse triples', len(inverse_triples))
        print('num equivalent triples', len(equivalent_triples))
        print('num subproperty triples', len(subproperty_triples))
        print('num inferencechain1 triples', len(inferencechain1_triples))
        print('num inferencechain2 triples', len(inferencechain2_triples))
        print('num inferencechain3 triples', len(inferencechain3_triples))
        print('num inferencechain4 triples', len(inferencechain4_triples))
        #print(self.train_ids_labels_inject)
        updated_train_data=self.generate_new_train_triples()
        return updated_train_data

    def split_embedding(self, embedding):
        """split embedding

        Args:
            embedding: embeddings need to be splited, shape:[None, dim].

        Returns:
            probability: The similrity between two matrices.
        """
        # embedding: [None, dim]
        assert self.args.emb_dim % 4 == 0
        num_scalar = self.args.emb_dim // 2
        num_block = self.args.emb_dim // 4
        embedding_scalar = embedding[:, 0:num_scalar]
        embedding_x = embedding[:, num_scalar:-num_block]
        embedding_y = embedding[:, -num_block:]
        return embedding_scalar, embedding_x, embedding_y

    
    # calculate the similrity between two matrices
    # head: [?, dim]
    # tail: [?, dim] or [1,dim]
    def sim(self, head=None, tail=None, arity=None):
        """calculate the similrity between two matrices

        Args:
            head: embeddings of head, shape:[batch_size, dim].
            tail: embeddings of tail, shape:[batch_size, dim] or [1, dim].
            arity: 1，2 or 3

        Returns:
            probability: The similrity between two matrices.

        """
        if arity == 1:
            A_scalar, A_x, A_y = self.split_embedding(head)
        elif arity == 2:
            M1_scalar, M1_x, M1_y = self.split_embedding(head[0])
            M2_scalar, M2_x, M2_y = self.split_embedding(head[1])
            A_scalar= M1_scalar * M2_scalar
            A_x = M1_x*M2_x - M1_y*M2_y
            A_y = M1_x*M2_y + M1_y*M2_x
        elif arity==3:
            M1_scalar, M1_x, M1_y = self.split_embedding(head[0])
            M2_scalar, M2_x, M2_y = self.split_embedding(head[1])
            M3_scalar, M3_x, M3_y = self.split_embedding(head[2])
            M1M2_scalar = M1_scalar * M2_scalar
            M1M2_x = M1_x * M2_x - M1_y * M2_y
            M1M2_y = M1_x * M2_y + M1_y * M2_x
            A_scalar = M1M2_scalar * M3_scalar
            A_x = M1M2_x * M3_x - M1M2_y * M3_y
            A_y = M1M2_x * M3_y + M1M2_y * M3_x
        else:
            raise NotImplemented
        B_scala, B_x, B_y = self.split_embedding(tail)

        similarity = torch.cat([(A_scalar - B_scala)**2, (A_x - B_x)**2, (A_x - B_x)**2, (A_y - B_y)**2, (A_y - B_y)**2 ], dim=1)
        similarity = torch.sqrt(torch.sum(similarity, dim=1))

        #recale the probability
        probability = (torch.max(similarity)-similarity)/(torch.max(similarity)-torch.min(similarity))

        return probability

    # generate a probality for each axiom in axiom pool
    def run_axiom_probability(self):
        """this function is used to generate a probality for each axiom in axiom pool

        """
        self.identity = torch.cat((torch.ones(int(self.args.emb_dim-self.args.emb_dim/4)),torch.zeros(int(self.args.emb_dim/4))),0).unsqueeze(0).cuda()
        
        if len(self.axiompool_reflexive) != 0: 
            index = torch.LongTensor(self.axiompool_reflexive).cuda()
            reflexive_embed = self.rel_emb(index)
            reflexive_prob = self.sim(head=reflexive_embed[:, 0, :], tail=self.identity, arity=1)
        else: 
            reflexive_prob = []

        if len(self.axiompool_symmetric) != 0: 
            index = torch.LongTensor(self.axiompool_symmetric).cuda()
            symmetric_embed = self.rel_emb(index)
            symmetric_prob = self.sim(head=[symmetric_embed[:, 0, :], symmetric_embed[:, 0, :]], tail=self.identity, arity=2)
            #symmetric_prob = sess.run(self.symmetric_probability, {self.symmetric_pool: self.axiompool_symmetric})
        else: 
            symmetric_prob = []

        if len(self.axiompool_transitive) != 0: 
            index = torch.LongTensor(self.axiompool_transitive).cuda()
            transitive_embed = self.rel_emb(index)
            transitive_prob = self.sim(head=[transitive_embed[:, 0, :], transitive_embed[:, 0, :]], tail=transitive_embed[:, 0, :], arity=2)
            #transitive_prob = sess.run(self.transitive_probability, {self.transitive_pool: self.axiompool_transitive})
        else: 
            transitive_prob = []

        if len(self.axiompool_inverse) != 0: 
            index = torch.LongTensor(self.axiompool_inverse).cuda()
            #inverse_prob = sess.run(self.inverse_probability, {self.inverse_pool: self.axiompool_inverse})
            inverse_embed = self.rel_emb(index)
            inverse_probability1 = self.sim(head=[inverse_embed[:, 0,:],inverse_embed[:, 1,:]], tail = self.identity, arity=2)
            inverse_probability2 = self.sim(head=[inverse_embed[:,1,:],inverse_embed[:, 0,:]], tail=self.identity, arity=2)
            inverse_prob = (inverse_probability1 + inverse_probability2)/2
        else: 
            inverse_prob = []

        if len(self.axiompool_subproperty) != 0: 
            index = torch.LongTensor(self.axiompool_subproperty).cuda()
            #subproperty_prob = sess.run(self.subproperty_probability, {self.subproperty_pool: self.axiompool_subproperty})
            subproperty_embed = self.rel_emb(index)
            subproperty_prob = self.sim(head=subproperty_embed[:, 0,:], tail=subproperty_embed[:, 1, :], arity=1)
        else: 
            subproperty_prob = []

        if len(self.axiompool_equivalent) != 0: 
            index = torch.LongTensor(self.axiompool_equivalent).cuda()
            #equivalent_prob = sess.run(self.equivalent_probability, {self.equivalent_pool: self.axiompool_equivalent})
            equivalent_embed = self.rel_emb(index)
            equivalent_prob = self.sim(head=equivalent_embed[:, 0,:], tail=equivalent_embed[:, 1,:], arity=1)
        else: 
            equivalent_prob = []

        if len(self.axiompool_inferencechain1) != 0:
            index = torch.LongTensor(self.axiompool_inferencechain1).cuda()
            inferencechain_embed = self.rel_emb(index) 
            inferencechain1_prob = self.sim(head=[inferencechain_embed[:, 1, :], inferencechain_embed[:, 0, :]], tail=inferencechain_embed[:, 2, :], arity=2)
        else:
            inferencechain1_prob = []

        if len(self.axiompool_inferencechain2) != 0:
            index = torch.LongTensor(self.axiompool_inferencechain2).cuda()
            inferencechain_embed = self.rel_emb(index)
            inferencechain2_prob = self.sim(head=[inferencechain_embed[:, 2, :], inferencechain_embed[:, 1, :], inferencechain_embed[:, 0, :]], tail=self.identity, arity=3)
        else:
            inferencechain2_prob = []

        if len(self.axiompool_inferencechain3) != 0:
            index = torch.LongTensor(self.axiompool_inferencechain3).cuda()
            inferencechain_embed = self.rel_emb(index)
            inferencechain3_prob = self.sim(head=[inferencechain_embed[:, 1, :], inferencechain_embed[:, 2, :]], tail=inferencechain_embed[:, 0, :], arity=2)
        else:
            inferencechain3_prob = []

        if len(self.axiompool_inferencechain4) != 0:
            index = torch.LongTensor(self.axiompool_inferencechain4).cuda()
            inferencechain_embed = self.rel_emb(index)
            inferencechain4_prob = self.sim(head=[inferencechain_embed[:, 0, :], inferencechain_embed[:, 2, :]],tail=inferencechain_embed[:, 1, :], arity=2)
        else:
            inferencechain4_prob = []

        output = [reflexive_prob, symmetric_prob, transitive_prob, inverse_prob,
                  subproperty_prob,equivalent_prob,inferencechain1_prob, inferencechain2_prob,
                  inferencechain3_prob, inferencechain4_prob]
        return output

    def update_valid_axioms(self, input):
        """this function is used to select high probability axioms as valid axioms and record their scores

        """
        # 
        # 
        valid_axioms = [self._select_high_probability(list(prob), axiom) for prob,axiom in zip(input, self.axiompool)]

        self.valid_reflexive, self.valid_symmetric, self.valid_transitive, \
        self.valid_inverse, self.valid_subproperty, self.valid_equivalent, \
        self.valid_inferencechain1, self.valid_inferencechain2, \
        self.valid_inferencechain3, self.valid_inferencechain4 = valid_axioms
        # update the batchsize of axioms and entailments
        self._reset_valid_axiom_entailment()


    def _select_high_probability(self, prob, axiom):
        # select the high probability axioms and recore their probabilities
        valid_axiom = [[axiom[prob.index(p)],[p]] for p in prob if p>self.select_probability]
        return valid_axiom

    def _reset_valid_axiom_entailment(self):

        self.infered_hr_t = defaultdict(set)
        self.infered_tr_h = defaultdict(set)

        self.valid_reflexive2entailment, self.valid_reflexive_p = \
            self._valid_axiom2entailment(self.valid_reflexive, self.reflexive2entailment)

        self.valid_symmetric2entailment, self.valid_symmetric_p = \
            self._valid_axiom2entailment(self.valid_symmetric, self.symmetric2entailment)

        self.valid_transitive2entailment, self.valid_transitive_p = \
            self._valid_axiom2entailment(self.valid_transitive, self.transitive2entailment)

        self.valid_inverse2entailment, self.valid_inverse_p = \
            self._valid_axiom2entailment(self.valid_inverse, self.inverse2entailment)

        self.valid_subproperty2entailment, self.valid_subproperty_p = \
            self._valid_axiom2entailment(self.valid_subproperty, self.subproperty2entailment)

        self.valid_equivalent2entailment, self.valid_equivalent_p = \
            self._valid_axiom2entailment(self.valid_equivalent, self.equivalent2entailment)


        self.valid_inferencechain12entailment, self.valid_inferencechain1_p = \
            self._valid_axiom2entailment(self.valid_inferencechain1, self.inferencechain12entailment)

        self.valid_inferencechain22entailment, self.valid_inferencechain2_p = \
            self._valid_axiom2entailment(self.valid_inferencechain2, self.inferencechain22entailment)

        self.valid_inferencechain32entailment, self.valid_inferencechain3_p = \
            self._valid_axiom2entailment(self.valid_inferencechain3, self.inferencechain32entailment)
        self.valid_inferencechain42entailment, self.valid_inferencechain4_p = \
            self._valid_axiom2entailment(self.valid_inferencechain4, self.inferencechain42entailment)


    def _valid_axiom2entailment(self, valid_axiom, axiom2entailment):
        valid_axiom2entailment = []
        valid_axiom_p = []
        for axiom_p in valid_axiom:
            axiom = tuple(axiom_p[0])
            p = axiom_p[1]
            for entailment in axiom2entailment[axiom]:
                valid_axiom2entailment.append(entailment)
                valid_axiom_p.append(p)
                h,r,t = entailment[-3:]
                self.infered_hr_t[(h,r)].add(t)
                self.infered_tr_h[(t,r)].add(h)
        return valid_axiom2entailment, valid_axiom_p


    # updata new train triples:
    def generate_new_train_triples(self):
        """The function is to updata new train triples and used after each training epoch end

        Returns:
            self.train_sampler.train_triples: The new training dataset (triples).
        """
        self.train_sampler.train_triples = copy.deepcopy(self.train_triples_base)
        print('generate_new_train_triples...')
        #origin_triples = train_sampler.train_triples
        inject_triples = self.train_ids_labels_inject
        inject_num = int(self.inject_triple_percent*len(self.train_sampler.train_triples))
        if len(inject_triples)> inject_num and inject_num >0:
            np.random.shuffle(inject_triples)
            inject_triples = inject_triples[:inject_num]
        #train_triples = np.concatenate([origin_triples, inject_triples], axis=0)
        print('当前train_sampler.train_triples数目',len(self.train_sampler.train_triples))
        
        for h,r,t in inject_triples:       
            self.train_sampler.train_triples.append((int(h),int(r),int(t)))
        print('添加后train_sampler.train_triples数目',len(self.train_sampler.train_triples))

        return self.train_sampler.train_triples
    
    def get_rule(self, rel2id):
        """Get rule for rule_base KGE models, such as ComplEx_NNE model.
        Get rule and confidence from _cons.txt file.
        Update:
            (rule_p, rule_q): Rule.
            confidence: The confidence of rule.
        """
        rule_p, rule_q, confidence = [], [], []
        with open(os.path.join(self.args.data_path, '_cons.txt')) as file:
            lines = file.readlines()
            for line in lines:
                rule_str, trust = line.strip().split()
                body, head = rule_str.split(',')
                if '-' in body:
                    rule_p.append(rel2id[body[1:]])
                    rule_q.append(rel2id[head])
                else:
                    rule_p.append(rel2id[body])
                    rule_q.append(rel2id[head])
                confidence.append(float(trust))
        rule_p = torch.tensor(rule_p).cuda()
        rule_q = torch.tensor(rule_q).cuda()
        confidence = torch.tensor(confidence).cuda()
        return (rule_p, rule_q), confidence

    """def init_emb(self):
        self.epsilon = 2.0
        self.margin = nn.Parameter(
            torch.Tensor([self.args.margin]), 
            requires_grad=False
        )
        self.embedding_range = nn.Parameter(
            torch.Tensor([(self.margin.item() + self.epsilon) / self.args.emb_dim]), 
            requires_grad=False
        )
        
        self.ent_emb = nn.Embedding(self.args.num_ent, self.args.emb_dim * 2)
        self.rel_emb = nn.Embedding(self.args.num_rel, self.args.emb_dim * 2)
        nn.init.uniform_(tensor=self.ent_emb.weight.data, a=-self.embedding_range.item(), b=self.embedding_range.item())
        nn.init.uniform_(tensor=self.rel_emb.weight.data, a=-self.embedding_range.item(), b=self.embedding_range.item())"""
    
    def init_emb(self):
        """Initialize the entity and relation embeddings in the form of a uniform distribution.

        """
        self.epsilon = 2.0
        self.margin = nn.Parameter(
            torch.Tensor([self.args.margin]), 
            requires_grad=False
        )
        self.embedding_range = nn.Parameter(
            torch.Tensor([(self.margin.item() + self.epsilon) / self.args.emb_dim]), 
            requires_grad=False
        )

        self.ent_emb = nn.Embedding(self.args.num_ent, self.args.emb_dim)
        self.rel_emb = nn.Embedding(self.args.num_rel, self.args.emb_dim)
        nn.init.uniform_(tensor=self.ent_emb.weight.data, a=-self.embedding_range.item(), b=self.embedding_range.item())
        nn.init.uniform_(tensor=self.rel_emb.weight.data, a=-self.embedding_range.item(), b=self.embedding_range.item())

        

    """def score_func(self, head_emb, relation_emb, tail_emb, mode):
        re_head, im_head = torch.chunk(head_emb, 2, dim=-1)
        re_relation, im_relation = torch.chunk(relation_emb, 2, dim=-1)
        re_tail, im_tail = torch.chunk(tail_emb, 2, dim=-1)

        return torch.sum(
            re_head * re_tail * re_relation
            + im_head * im_tail * re_relation
            + re_head * im_tail * im_relation
            - im_head * re_tail * im_relation,
            -1
        )"""
        
    def score_func(self, head_emb, relation_emb, tail_emb, mode):
        """Calculating the score of triples.
        
        The formula for calculating the score is DistMult.

        Args:
            head_emb: The head entity embedding.
            relation_emb: The relation embedding.
            tail_emb: The tail entity embedding.
            mode: Choose head-predict or tail-predict.

        Returns:
            score: The score of triples.
        """
        if mode == 'head-batch':
            score = head_emb * (relation_emb * tail_emb)
        else:
            score = (head_emb * relation_emb) * tail_emb

        score = score.sum(dim = -1)
        return score

    def forward(self, triples, negs=None, mode='single'):
        """The functions used in the training phase

        Args:
            triples: The triples ids, as (h, r, t), shape:[batch_size, 3].
            negs: Negative samples, defaults to None.
            mode: Choose head-predict or tail-predict, Defaults to 'single'.

        Returns:
            score: The score of triples.
        """
        head_emb, relation_emb, tail_emb = self.tri2emb(triples, negs, mode)
        score = self.score_func(head_emb, relation_emb, tail_emb, mode)

        return score
    
    def get_score(self, batch, mode):
        """The functions used in the testing phase

        Args:
            batch: A batch of data.
            mode: Choose head-predict or tail-predict.

        Returns:
            score: The score of triples.
        """
        triples = batch["positive_sample"]
        head_emb, relation_emb, tail_emb = self.tri2emb(triples, mode=mode)
        score = self.score_func(head_emb, relation_emb, tail_emb, mode)
        return score



import torch
import torch.nn.functional as F
import torch.nn as nn
from IPython import embed

class SimplE_Loss(nn.Module):
    def __init__(self, args, model):
        super(SimplE_Loss, self).__init__()
        self.args = args
        self.model = model
    def forward(self, pos_sample, neg_sample, mode):
        neg_score = self.model(pos_sample, neg_sample, mode) #shape:[bs, neg_num]
        pos_score = -self.model(pos_sample).view(-1, 1) #shape:[bs]
        score = torch.cat((neg_score, pos_score), dim = -1) #shape:[bs, neg_num+1]
        loss = torch.sum(F.softplus(score)) + self.args.regularization * self.model.l2_loss()

        return loss


        



import torch.nn as nn
import torch
from IPython import embed
from neuralkg.model.KGEModel.model import Model

class XTransE(Model):

    """`Explainable Knowledge Graph Embedding for Link Prediction with Lifestyles in e-Commerce`_ (XTransE), which introduces the attention to aggregate the neighbor node representation.

    Attributes:
        args: Model configuration parameters.
    
    .. _Explainable Knowledge Graph Embedding for Link Prediction with Lifestyles in e-Commerce: https://link.springer.com/content/pdf/10.1007%2F978-981-15-3412-6_8.pdf
    """

    def __init__(self, args):
        super(XTransE, self).__init__(args)
        self.args = args
        self.ent_emb = None
        self.rel_emb = None

        self.init_emb()

    def init_emb(self):

        """Initialize the entity and relation embeddings in the form of a uniform distribution.

        Args:
            margin: Caculate embedding_range and loss.
            embedding_range: Uniform distribution range.
            ent_emb: Entity embedding, shape:[num_ent, emb_dim].
            rel_emb: Relation_embedding, shape:[num_rel, emb_dim].
        """

        self.margin = nn.Parameter(
            torch.Tensor([self.args.margin]), 
            requires_grad=False
        )
        self.embedding_range = nn.Parameter(
            torch.Tensor([6.0 / float(self.args.emb_dim).__pow__(0.5)]),
            requires_grad=False
        )
        self.ent_emb = nn.Embedding(self.args.num_ent, self.args.emb_dim)
        self.rel_emb = nn.Embedding(self.args.num_rel, self.args.emb_dim)
        nn.init.uniform_(tensor=self.ent_emb.weight.data, a=-self.embedding_range.item(), b=self.embedding_range.item())
        nn.init.uniform_(tensor=self.rel_emb.weight.data, a=-self.embedding_range.item(), b=self.embedding_range.item())

    def score_func(self, triples, neighbor=None, mask=None, negs=None, mode='single'):

        """Calculating the score of triples.
        
        Args:
            triples: The triples ids, as (h, r, t), shape:[batch_size, 3].
            neighbor: The neighbors of tail entities.
            mask: The mask of neighbor nodes
            negs: Negative samples, defaults to None.
            mode: Choose head-predict or tail-predict, Defaults to 'single'.                                                                                                                                                                            

        Returns:
            score: The score of triples.
        """

        head = triples[:,0]
        rela = triples[:,1]
        tail = triples[:,2]

        if mode == 'tail-batch':
            tail = negs.squeeze(1)
        
        norm_emb_ent = nn.functional.normalize(self.ent_emb.weight, dim=1, p=2) # [ent, dim]
        norm_emb_rel = nn.functional.normalize(self.rel_emb.weight, dim=1, p=2) # [rel, dim]
        
        neighbor_tail_emb = norm_emb_ent[neighbor[:, :, 1]] # [batch, neighbor, dim]
        neighbor_rela_emb = norm_emb_rel[neighbor[:, :, 0]] # [batch, neighbor, dim]
        neighbor_head_emb = neighbor_tail_emb - neighbor_rela_emb 

        rela_emb = norm_emb_rel[rela] # [batch, dim]
        tail_emb = norm_emb_ent[tail] # [batch, dim]
        head_emb = norm_emb_ent[head]
        h_rt_embedding = tail_emb - rela_emb

        attention_rt = torch.zeros([self.args.train_bs, 200]).type_as(self.ent_emb.weight)
        attention_rt = (neighbor_head_emb * h_rt_embedding.unsqueeze(1)).sum(dim=2) * mask
        attention_rt = nn.functional.softmax(attention_rt, dim=1).unsqueeze(2)

        head_emb = head_emb + \
            torch.bmm(neighbor_head_emb.permute(0,2,1), attention_rt).reshape([-1,self.args.emb_dim])

        score = self.margin.item() - torch.norm(head_emb + rela_emb - tail_emb, p=2, dim=1)
        return score.unsqueeze(1)

    def transe_func(self, head_emb, rela_emb, tail_emb):
        
        """Calculating the score of triples with TransE model.

        Args:
            head_emb: The head entity embedding.
            rela_emb: The relation embedding.
            tail_emb: The tail entity embedding.

        Returns:
            score: The score of triples.
        """

        score = (head_emb + rela_emb) - tail_emb
        score = self.margin.item() - torch.norm(score, p=2, dim=-1)
        return score

    def forward(self, triples, neighbor=None, mask=None, negs=None, mode='single'):

        """The functions used in the training and testing phase

        Args:
            triples: The triples ids, as (h, r, t), shape:[batch_size, 3].
            neighbor: The neighbors of tail entities.
            mask: The mask of neighbor nodes
            negs: Negative samples, defaults to None.
            mode: Choose head-predict or tail-predict, Defaults to 'single'.

        Returns:
            score: The score of triples.
        """

        head_emb, relation_emb, tail_emb = self.tri2emb(triples, negs, mode)
        TransE_score = self.transe_func(head_emb, relation_emb, tail_emb)
        XTransE_score = self.score_func(triples, neighbor, mask, negs, mode)
        
        return TransE_score + XTransE_score






import torch
import torch.nn.functional as F
import torch.nn as nn

class RGCN_Loss(nn.Module):

    def __init__(self, args, model):
        super(RGCN_Loss, self).__init__()
        self.args = args
        self.model = model
    
    def reg_loss(self): 
        return torch.mean(self.model.Loss_emb.pow(2)) + torch.mean(self.model.rel_emb.pow(2))

    def forward(self, score, labels):
         loss = F.binary_cross_entropy_with_logits(score, labels)
         regu = self.args.regularization * self.reg_loss()
         loss += regu
         return loss